from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'aggregate': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        db = query['$db']
        collection = query['aggregate']

        count = 0
        if db == 'mindsdb' and collection == 'predictors':
            count = len(mindsdb_env['model_interface'].get_models())

        return {
            'count': count,
            'maxTimeMs': 5000,
            '$db': db
        }


responder = Responce()



from .whatsmyuri import responder as responder_whatsmyuri
from .buildinfo import responder as responder_buildinfo
from .is_master import responder as responder_is_master
from .is_master_lower import responder as responder_is_master_lower
from .replsetgetstatus import responder as responder_replsetgetstatus
from .getlog import responder as responder_getlog
from .add_shard import responder as responder_add_shard
from .update_range_deletions import responder as responder_update_range_deletions
from .recv_chunk_start import responder as responder_recv_chunk_start
from .connection_status import responder as responder_connection_status
from .get_cmd_line_opts import responder as responder_get_cmd_line_opts
from .host_info import responder as responder_host_info
from .db_stats import responder as responder_db_stats
from .coll_stats import responder as responder_coll_stats
from .count import responder as responder_count
from .aggregate import responder as responder_aggregate
from .get_free_monitoring_status import responder as responder_get_free_monitoring_status
from .end_sessions import responder as responder_end_sessions
from .ping import responder as responder_ping

from .list_indexes import responder as responder_list_indexes
from .list_collections import responder as responder_list_collections
from .list_databases import responder as responder_list_databases

from .find import responder as responder_find
from .insert import responder as responder_insert
from .delete import responder as responder_delete

from .sasl_start import responder as sasl_start
from .sasl_continue import responder as sasl_continue

from .company_id import responder as responder_company_id


responders = [
    # service queries
    responder_whatsmyuri,
    responder_buildinfo,
    responder_is_master,
    responder_is_master_lower,
    responder_replsetgetstatus,
    responder_getlog,
    responder_add_shard,                # 4.4
    responder_update_range_deletions,   # 4.4
    responder_recv_chunk_start,         # 4.4
    responder_connection_status,
    responder_get_cmd_line_opts,
    responder_host_info,
    responder_db_stats,
    responder_coll_stats,
    responder_count,
    responder_aggregate,
    responder_get_free_monitoring_status,
    responder_end_sessions,
    responder_ping,
    # user queries
    responder_list_indexes,
    responder_list_collections,
    responder_list_databases,
    responder_find,
    responder_insert,
    responder_delete,
    # auth
    sasl_start,
    sasl_continue,
    # cloud
    responder_company_id
]



from pandas import DataFrame

from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE

from mindsdb_sql.parser.ast import ASTNode

class HandlerResponse:
    def __init__(self, resp_type: RESPONSE_TYPE, data_frame: DataFrame = None,
                 query: ASTNode = 0, error_code: int = 0, error_message: str = None) -> None:
        self.resp_type = resp_type
        self.query = query
        self.data_frame = data_frame
        self.error_code = error_code
        self.error_message = error_message

    @property
    def type(self):
        return self.resp_type


class HandlerStatusResponse:
    def __init__(self, success: bool = True, error_message: str = None) -> None:
        self.success = success
        self.error_message = error_message



import unittest

from mindsdb.integrations.mysql_handler.mysql_handler import MySQLHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class MySQLHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "localhost",
            "port": "3306",
            "user": "root",
            "password": "",
            "database": "test",
            "ssl": False
        }
        cls.handler = MySQLHandler('test_singlestore_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.check_connection()

    def test_1_native_query_show_dbs(self):
        dbs = self.handler.native_query("SHOW DATABASES;")
        assert dbs['type'] is not RESPONSE_TYPE.ERROR

    def test_2_get_tables(self):
        tbls = self.handler.get_tables()
        assert tbls['type'] is not RESPONSE_TYPE.ERROR

    def test_3_get_views(self):
        views = self.handler.get_views()
        assert views['type'] is not RESPONSE_TYPE.ERROR
    
    def test_5_drop_table(self):
        res = self.handler.native_query("DROP TABLE IF EXISTS test_mdb")
        assert res['type'] is not RESPONSE_TYPE.ERROR 

    def test_4_create_table(self):
        res = self.handler.native_query("CREATE TABLE IF NOT EXISTS test_mdb (test_col INT)")
        assert res['type'] is not RESPONSE_TYPE.ERROR 

    def test_6_describe_table(self):
        described = self.handler.describe_table("test_mdb")
        assert described['type'] is RESPONSE_TYPE.TABLE

    def test_7_select_query(self):
        query = "SELECT * FROM test_mdb WHERE 'id'='a'"
        result = self.handler.select_query(query)
        assert result['type'] is RESPONSE_TYPE.TABLE


"""
*******************************************************
 * Copyright (C) 2017 MindsDB Inc. <copyright@mindsdb.com>
 *
 * This file is part of MindsDB Server.
 *
 * MindsDB Server can not be copied and/or distributed without the express
 * permission of MindsDB Inc
 *******************************************************
"""

import struct
import math

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import Parameter

from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import COMMANDS, getConstName, TYPES


class CommandPacket(Packet):
    '''
    Implementation based on description:
    https://mariadb.com/kb/en/library/1-connecting-connecting/#initial-handshake-packet
    '''

    def _read_byte(self, buffer):
        b = buffer[:1]
        buffer = buffer[1:]
        b = struct.unpack('<B', b)[0]
        return b, buffer

    def read_params(self, buffer, num_params):
        if not num_params > 0:
            return

        # read null-map
        null_bytes = math.floor((num_params + 7) / 8)
        nulls = []
        for i in range(null_bytes):
            b, buffer = self._read_byte(buffer)
            for i in range(8):
                nulls.append(((1 << i) & b) != 0)

        # read send-type byte
        b, buffer = self._read_byte(buffer)

        types = []
        if b == 1:
            # read types
            for i in range(num_params):
                t, buffer = self._read_byte(buffer)
                s, buffer = self._read_byte(buffer)
                types.append(dict(
                    type=t,
                    signed=s
                ))

        datumtypes = {
            TYPES.MYSQL_TYPE_VAR_STRING: 'string<lenenc>',
            TYPES.MYSQL_TYPE_STRING: 'string<lenenc>',
            TYPES.MYSQL_TYPE_VARCHAR: 'string<lenenc>',

            TYPES.MYSQL_TYPE_TINY: 'int<1>',
            TYPES.MYSQL_TYPE_SHORT: 'int<2>',
            TYPES.MYSQL_TYPE_LONG: 'int<4>',
            TYPES.MYSQL_TYPE_LONGLONG: 'int<8>',
        }

        for i in range(num_params):
            if nulls[i]:
                self.parameters.append(None)
                continue

            datum_type = datumtypes.get(types[i]['type'])
            if datum_type is not None:
                x = Datum(datum_type)
                buffer = x.setFromBuff(buffer)
                value = x.value
                if isinstance(value, bytes):
                    value = value.decode()

                self.parameters.append(value)
            else:
                # NOTE at this moment all sends as strings and it works
                raise Exception(f"Unsupported type {types[i]['type']}")

    def setup(self, length=0, count_header=1, body=''):
        if length == 0:
            return

        # self.salt=self.session.salt

        self._length = length
        self._seq = count_header
        self._body = body

        self.type = Datum('int<1>')
        buffer = body
        buffer = self.type.setFromBuff(buffer)

        if self.type.value in (COMMANDS.COM_QUERY, COMMANDS.COM_STMT_PREPARE):
            self.sql = Datum('str<EOF>')
            buffer = self.sql.setFromBuff(buffer)
        elif self.type.value == COMMANDS.COM_STMT_EXECUTE:
            # https://mariadb.com/kb/en/com_stmt_execute/
            self.stmt_id = Datum('int<4>')
            buffer = self.stmt_id.setFromBuff(buffer)
            self.flags = Datum('int<1>')
            buffer = self.flags.setFromBuff(buffer)
            self.iteration_count = Datum('int<4>')
            buffer = self.iteration_count.setFromBuff(buffer)

            self.parameters = []

            prepared_stmt = self.session.prepared_stmts[self.stmt_id.value]

            num_params = len(prepared_stmt['statement'].params)
            self.read_params(buffer, num_params)
            #
            # if prepared_stmt['type'] == 'select':
            #     num_params = len(prepared_stmt['statement'].parameters)
            #
            #     self.read_params(buffer, num_params)
            #
            # elif prepared_stmt['type'] in ['insert', 'delete']:
            #     # if prepared_stmt['type'] == 'insert':
            #     #     prepared_stmt['statement'].sql
            #     #     statement = parse_sql(prepared_stmt['statement'].sql)
            #     #     num_params = 0
            #     #     for row in statement.values:
            #     #         for item in row:
            #     #             if isinstance(item, Parameter):
            #     #                 num_params = num_params + 1
            #     # elif prepared_stmt['type'] == 'delete':
            #     #     num_params = prepared_stmt['statement'].sql.count('?')
            #
            #     num_params = len(prepared_stmt['statement'].parameters)
            #     self.read_params(buffer, num_params)


        elif self.type.value == COMMANDS.COM_STMT_CLOSE:
            self.stmt_id = Datum('int<4>')
            buffer = self.stmt_id.setFromBuff(buffer)
        elif self.type.value == COMMANDS.COM_STMT_FETCH:
            self.stmt_id = Datum('int<4>')
            buffer = self.stmt_id.setFromBuff(buffer)
            self.limit = Datum('int<4>')
            buffer = self.limit.setFromBuff(buffer)
        elif self.type.value == COMMANDS.COM_INIT_DB:
            self.database = Datum('str<EOF>')
            buffer = self.database.setFromBuff(buffer)
        else:
            self.data = Datum('str<EOF>')
            buffer = self.data.setFromBuff(buffer)

    def __str__(self):
        return str({
            'header': {'length': self.length, 'seq': self.seq},
            'type': getConstName(COMMANDS, self.type.value),
            'vars': self.__dict__
        })



from mindsdb_sql.exceptions import PlanningException
from mindsdb_sql.planner.utils import get_integration_path_from_identifier
from mindsdb_sql.parser.ast import Identifier, Operation, BinaryOperation, BetweenOperation, NullConstant, Latest, Select, Star, Constant


def get_time_selects(time_filter, table, window, order_by, preparation_where):
    # TODO: add tests for all these cases
    time_col = str(order_by[0].field)
    if isinstance(time_filter, BetweenOperation):
        between_from = time_filter.args[1]
        preparation_time_filter = BinaryOperation('<', args=[Identifier(time_col), between_from])
        preparation_where2 = replace_time_filter(preparation_where2, time_filter, preparation_time_filter)
        integration_select_1 = Select(targets=[Star()],
                                      from_table=table,
                                      where=add_order_not_null(preparation_where2),
                                      order_by=order_by,
                                      limit=Constant(window))

        integration_select_2 = Select(targets=[Star()],
                                      from_table=table,
                                      where=preparation_where,
                                      order_by=order_by)

        integration_selects = [integration_select_1, integration_select_2]
    elif isinstance(time_filter, BinaryOperation) and time_filter.op == '>' and time_filter.args[1] == Latest():
        integration_select = Select(targets=[Star()],
                                    from_table=table,
                                    where=preparation_where,
                                    order_by=order_by,
                                    limit=Constant(window),
                                    )
        integration_select.where = find_and_remove_time_filter(integration_select.where, time_filter)
        integration_selects = [integration_select]

    elif isinstance(time_filter, BinaryOperation) and time_filter.op in ('>', '>='):
        time_filter_date = time_filter.args[1]
        preparation_time_filter_op = {'>': '<=', '>=': '<'}[time_filter.op]

        preparation_time_filter = BinaryOperation(preparation_time_filter_op,
                                                  args=[Identifier(time_col), time_filter_date])
        preparation_where2 = replace_time_filter(preparation_where2, time_filter, preparation_time_filter)
        integration_select_1 = Select(targets=[Star()],
                                      from_table=table,
                                      where=add_order_not_null(preparation_where2),
                                      order_by=order_by,
                                      limit=Constant(window))

        integration_select_2 = Select(targets=[Star()],
                                      from_table=table,
                                      where=preparation_where,
                                      order_by=order_by)

        integration_selects = [integration_select_1, integration_select_2]
    else:
        integration_select = Select(targets=[Star()],
                                    from_table=table,
                                    where=preparation_where,
                                    order_by=order_by,
                                    )
        integration_selects = [integration_select]

    return integration_selects


def find_time_filter(op, time_column_name):
    if not op:
        return
    if op.op == 'and':
        left = find_time_filter(op.args[0], time_column_name)
        right = find_time_filter(op.args[1], time_column_name)
        if left and right:
            raise PlanningException('Can provide only one filter by predictor order_by column, found two')

        return left or right
    elif ((isinstance(op.args[0], Identifier) and op.args[0].parts[-1].lower() == time_column_name.lower()) or
          (isinstance(op.args[1], Identifier) and op.args[1].parts[-1].lower() == time_column_name.lower())):
        return op


def replace_time_filter(op, time_filter, new_filter):
    if op == time_filter:
        return new_filter
    elif op.args[0] == time_filter:
        op.args[0] = new_filter
    elif op.args[1] == time_filter:
        op.args[1] = new_filter
    return op


def add_order_not_null(condition, time_column_name):
    order_field_not_null = BinaryOperation(op='is not', args=[
        Identifier(parts=[time_column_name]),
        NullConstant()
    ])
    if condition is not None:
        condition = BinaryOperation(op='and', args=[
            condition,
            order_field_not_null
        ])
    else:
        condition = order_field_not_null
    return condition


def find_and_remove_time_filter(op, time_filter):
    if isinstance(op, BinaryOperation) or isinstance(op, BetweenOperation):
        if op == time_filter:
            return None
        elif op.op == 'and':
            # TODO maybe OR operation too?

            # next level
            left_arg = find_and_remove_time_filter(op.args[0], time_filter)
            right_arg = find_and_remove_time_filter(op.args[1], time_filter)

            # if found in one arg return other
            if left_arg is None:
                return right_arg
            if right_arg is None:
                return left_arg

            op.args = [left_arg, right_arg]
            return op

    return op


def validate_ts_where_condition(op, allowed_columns, allow_and=True):
    """Error if the where condition caontains invalid ops, is nested or filters on some column that's not time or partition"""
    if not op:
        return
    allowed_ops = ['and', '>', '>=', '=', '<', '<=', 'between', 'in']
    if not allow_and:
        allowed_ops.remove('and')
    if op.op not in allowed_ops:
        raise PlanningException(
            f'For time series predictors only the following operations are allowed in WHERE: {str(allowed_ops)}, found instead: {str(op)}.')

    for arg in op.args:
        if isinstance(arg, Identifier):
            if arg.parts[-1].lower() not in allowed_columns:
                raise PlanningException(
                    f'For time series predictor only the following columns are allowed in WHERE: {str(allowed_columns)}, found instead: {str(arg)}.')

    if isinstance(op.args[0], Operation):
        validate_ts_where_condition(op.args[0], allowed_columns, allow_and=True)
    if isinstance(op.args[1], Operation):
        validate_ts_where_condition(op.args[1], allowed_columns, allow_and=True)



import os
import psutil
import tempfile
from pathlib import Path

from flask import request
from flask_restx import Resource
from flask import current_app as ca

from mindsdb.utilities.log import log
from mindsdb.api.http.namespaces.configs.util import ns_conf
from mindsdb.utilities.telemetry import (
    enable_telemetry,
    disable_telemetry,
    telemetry_file_exists,
    inject_telemetry_to_static
)
from mindsdb.api.http.initialize import update_static


@ns_conf.route('/ping')
class Ping(Resource):
    @ns_conf.doc('get_ping')
    def get(self):
        '''Checks server avaliable'''
        return {'status': 'ok'}


@ns_conf.route('/ping_native')
class PingNative(Resource):
    @ns_conf.doc('get_ping_native')
    def get(self):
        ''' Checks server use native for learn or analyse.
            Will return right result only on Linux.
        '''
        response = {
            'learn': False,
            'predict': False,
            'analyse': False
        }

        if os.name != 'posix':
            return response

        for process_type in response:
            processes_dir = Path(tempfile.gettempdir()).joinpath(f'mindsdb/processes/{process_type}/')
            if not processes_dir.is_dir():
                continue
            process_marks = [x.name for x in processes_dir.iterdir()]
            for p_mark in process_marks:
                pid = int(p_mark.split('-')[0])
                try:
                    psutil.Process(pid)
                except Exception:
                    processes_dir.joinpath(p_mark).unlink()
                else:
                    response[process_type] = True

        return response


@ns_conf.route('/telemetry')
class Telemetry(Resource):
    @ns_conf.doc('get_telemetry_status')
    def get(self):
        storage_dir = ca.config_obj['storage_dir']
        status = "enabled" if telemetry_file_exists(storage_dir) else "disabled"
        return {"status": status}

    @ns_conf.doc('set_telemetry')
    def post(self):
        data = request.json
        action = data['action']
        if str(action).lower() in ["true", "enable", "on"]:
            enable_telemetry(ca.config_obj['storage_dir'])
        else:
            disable_telemetry(ca.config_obj['storage_dir'])
        inject_telemetry_to_static(ca.config_obj.paths['static'])
        return '', 200

@ns_conf.route('/validate_json_ai')
class ValidateJsonAI(Resource):
    def post(self):
        json_ai = request.json.get('json_ai')
        if json_ai is None:
            return 'Please provide json_ai', 400
        try:
            code = request.model_interface.code_from_json_ai(json_ai)
        except Exception as e:
            return {'error': str(e)}
        return {'code': code}

@ns_conf.route('/update-gui')
class UpdateGui(Resource):
    @ns_conf.doc('get_update_gui')
    def get(self):
        update_static()
        return '', 200



import os
import logging
import torch.multiprocessing as mp
import threading
from pathlib import Path

from werkzeug.exceptions import HTTPException
from waitress import serve
from flask import send_from_directory, request, current_app
from flask_compress import Compress

from mindsdb.api.http.namespaces.predictor import ns_conf as predictor_ns
from mindsdb.api.http.namespaces.stream import ns_conf as stream_ns
from mindsdb.api.http.namespaces.config import ns_conf as conf_ns
from mindsdb.api.http.namespaces.util import ns_conf as utils_ns
from mindsdb.api.http.namespaces.file import ns_conf as file_ns
from mindsdb.api.http.namespaces.sql import ns_conf as sql_ns
from mindsdb.api.http.namespaces.analysis import ns_conf as analysis_ns
from mindsdb.api.http.namespaces.handlers import ns_conf as handlers_ns
from mindsdb.api.nlp.nlp import ns_conf as nlp_ns
from mindsdb.api.http.initialize import initialize_flask, initialize_interfaces, initialize_static
from mindsdb.utilities.with_kwargs_wrapper import WithKWArgsWrapper
from mindsdb.utilities.log import initialize_log, get_log
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage.db import session, engine as db_engine


def start(verbose, no_studio, with_nlp):
    config = Config()

    server = os.environ.get('MINDSDB_DEFAULT_SERVER', 'waitress')

    initialize_log(config, 'http', wrap_print=True if server.lower() != 'gunicorn' else False)

    # start static initialization in a separate thread
    init_static_thread = None
    if not no_studio:
        init_static_thread = threading.Thread(target=initialize_static)
        init_static_thread.start()

    app, api = initialize_flask(config, init_static_thread, no_studio)
    Compress(app)
    initialize_interfaces(app)

    static_root = config['paths']['static']
    if os.path.isabs(static_root) is False:
        static_root = os.path.join(os.getcwd(), static_root)
    static_root = Path(static_root)

    @app.route('/', defaults={'path': ''}, methods=['GET'])
    @app.route('/<path:path>', methods=['GET'])
    def root_index(path):
        if path.startswith('api/'):
            return {'message': 'wrong query'}, 400
        if static_root.joinpath(path).is_file():
            return send_from_directory(static_root, path)
        else:
            return send_from_directory(static_root, 'index.html')

    api.add_namespace(predictor_ns)
    api.add_namespace(stream_ns)
    api.add_namespace(utils_ns)
    api.add_namespace(conf_ns)
    api.add_namespace(file_ns)
    api.add_namespace(sql_ns)
    api.add_namespace(analysis_ns)
    api.add_namespace(handlers_ns)
    if with_nlp:
        api.add_namespace(nlp_ns)

    @api.errorhandler(Exception)
    def handle_exception(e):
        get_log('http').error(f'http exception: {e}')
        # pass through HTTP errors
        if isinstance(e, HTTPException):
            return {'message': str(e)}, e.code, e.get_response().headers
        name = getattr(type(e), '__name__') or 'Unknown error'
        return {'message': f'{name}: {str(e)}'}, 500

    @app.teardown_appcontext
    def remove_session(*args, **kwargs):
        session.close()

    @app.before_request
    def before_request():
        company_id = request.headers.get('company-id')
        user_class = request.headers.get('user-class')

        if company_id is not None:
            try:
                company_id = int(company_id)
            except Exception as e:
                get_log('http').error(f'Cloud not parse company id: {company_id} | exception: {e}')
                company_id = None

        if user_class is not None:
            try:
                user_class = int(user_class)
            except Exception as e:
                get_log('http').error(f'Cloud not parse user_class: {user_class} | exception: {e}')
                user_class = 0
        else:
            user_class = 0

        request.company_id = company_id
        request.user_class = user_class

        request.model_interface = WithKWArgsWrapper(
            current_app.original_model_interface,
            company_id=company_id
        )

        request.integration_controller = WithKWArgsWrapper(
            current_app.original_integration_controller,
            company_id=company_id
        )

        request.file_controller = WithKWArgsWrapper(
            current_app.original_file_controller,
            company_id=company_id
        )

    port = config['api']['http']['port']
    host = config['api']['http']['host']

    # waiting static initialization
    if not no_studio:
        init_static_thread.join()
    if server.lower() == 'waitress':
        if host in ('', '0.0.0.0'):
            serve(app, port=port, host='*', max_request_body_size=1073741824 * 10, inbuf_overflow=1073741824 * 10)
        else:
            serve(app, port=port, host=host, max_request_body_size=1073741824 * 10, inbuf_overflow=1073741824 * 10)
    elif server.lower() == 'flask':
        # that will 'disable access' log in console
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.WARNING)

        app.run(debug=False, port=port, host=host)
    elif server.lower() == 'gunicorn':
        try:
            from mindsdb.api.http.gunicorn_wrapper import StandaloneApplication
        except ImportError:
            print("Gunicorn server is not available by default. If you wish to use it, please install 'gunicorn'")
            return

        def post_fork(arbiter, worker):
            db_engine.dispose()

        options = {
            'bind': f'{host}:{port}',
            'workers': mp.cpu_count(),
            'timeout': 600,
            'reuse_port': True,
            'preload_app': True,
            'post_fork': post_fork,
            'threads': 4
        }
        StandaloneApplication(app, options).run()



from mindsdb.api.mysql.mysql_proxy.data_types.mysql_packet import Packet
from mindsdb.api.mysql.mysql_proxy.data_types.mysql_datum import Datum

class FastAuthFail(Packet):
    def setup(self):
        self.cont = Datum('int<1>', 4) # 0x04

    @property
    def body(self):

        order = [
            'cont'
        ]
        string = b''
        for key in order:
            string += getattr(self, key).toStringPacket()

        self.setBody(string)
        return self._body



import requests

from typing import Dict, List, Optional
from datetime import datetime

from mindsdb.integrations.libs.base_handler import BaseHandler, PredictiveHandler
from mindsdb.integrations.libs.storage_handler import SqliteStorageHandler
from mindsdb.utilities.config import Config
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import Join
from mindsdb_sql.parser.dialects.mindsdb import (
    # RetrainPredictor,
    CreatePredictor,
    DropPredictor,
)

import mlflow
from mlflow.tracking import MlflowClient
import pandas as pd


class MLflowHandler(PredictiveHandler):

    type = 'mlflow'

    def __init__(self, name):
        """
        An MLflow integration needs to have a working connection to work. For this:
            - All models to use should be previously served
            - An mlflow server should be running, to access the model registry
            
        Example:
            1. Run `mlflow server -p 5001 --backend-store-uri sqlite:///mlflow.db --default-artifact-root ./artifacts --host 0.0.0.0`
            2. Run `mlflow models serve --model-uri ./model_path`
            3. Instance this integration and call the `connect method` passing the relevant urls to mlflow and to the DB
            
        Note: above, `artifacts` is a folder to store artifacts for new experiments that do not specify an artifact store.
        """  # noqa
        super().__init__(name)
        self.mlflow_server_url = None
        self.mlflow_server_path = None
        self.connection = None
        self.storage = None
        self.parser = parse_sql
        self.dialect = 'mindsdb'

    def _check_model_url(self, url):
        # try to post without data and check status code not in (not_found, method_not_allowed)
        try:
            resp = requests.post(url)
            if resp.status_code in (404, 405):
                raise Exception(f'Model url is incorrect, status_code: {resp.status_code}')
        except requests.RequestException as e:
            raise Exception(f'Model url is incorrect: {str(e)}')

    def connect(self, **kwargs) -> Dict[str, int]:
        """ Connect to the mlflow process using MlflowClient class. """  # noqa
        self.mlflow_server_url = kwargs['mlflow_server_url']
        self.mlflow_server_path = kwargs['model_registry_path']
        self.connection = MlflowClient(self.mlflow_server_url, self.mlflow_server_path)
        self.storage = SqliteStorageHandler(context=self.name, config=kwargs['config'])
        return self.check_connection()

    def check_connection(self) -> Dict[str, int]:
        """ Checks that the connection is, as expected, an MlflowClient instance. """  # noqa
        # todo: as it stands this does not truly check if the connection is alive...
        try:
            assert isinstance(self.connection, mlflow.tracking.MlflowClient)
        except AssertionError as e:
            return {'status': '503', 'error': e}
        return {'status': '200'}

    def get_tables(self) -> List:
        """ Returns list of model names (that have been succesfully linked with CREATE PREDICTOR) """  # noqa
        models = self.storage.get('models')
        return list(models.keys()) if models else []

    def describe_table(self, table_name: str) -> Dict:
        """ For getting standard info about a table. e.g. data types """  # noqa
        if table_name not in self.get_tables():
            print("Table not found.")
            return {}

        models = {model.name: model for model in self.connection.list_registered_models()}
        model = models[table_name]
        latest_version = model.latest_versions[-1]
        description = {
            'NAME': model.name,
            'USER_DESCRIPTION': model.description,
            'LAST_STATUS': latest_version.status,
            'CREATED_AT': datetime.fromtimestamp(model.creation_timestamp//1000).strftime("%m/%d/%Y, %H:%M:%S"),
            'LAST_UPDATED': datetime.fromtimestamp(model.last_updated_timestamp//1000).strftime("%m/%d/%Y, %H:%M:%S"),
            'TAGS': model.tags,
            'LAST_RUN_ID': latest_version.run_id,
            'LAST_SOURCE_PATH': latest_version.source,
            'LAST_USER_ID': latest_version.user_id,
            'LAST_VERSION': latest_version.version,
        }
        return description

    def run_native_query(self, query_str: str) -> Optional[object]:
        statement = self.parser(query_str, dialect=self.dialect)

        if type(statement) == CreatePredictor:
            model_name = statement.name.parts[-1]

            # check that it exists within mlflow and is not already registered
            mlflow_models = [model.name for model in self.connection.list_registered_models()]
            if model_name not in mlflow_models:
                print("Error: this predictor is not registered in mlflow. Check you are serving it and try again.")
            elif model_name in self.get_tables():
                # @TODO: maybe add re-wiring so that a predictor name can point to a new endpoint?
                # e.g. add _edit_invocation_url method that edits the db.record and self.internal_registry
                print("Error: this model is already registered!")
            else:
                target = statement.targets[0].parts[-1]  # TODO: multiple target support?
                url = statement.using['url.predict']
                if url is not None:
                    self._check_model_url(url)
                params = {
                    'target': target,
                    'url': statement.using['url.predict']
                }
                all_models = self.storage.get('models')
                if all_models is not None:
                    all_models[model_name] = params
                else:
                    all_models = {model_name: params}
                self.storage.set('models', all_models)

        elif type(statement) == DropPredictor:
            to_drop = statement.name.parts[-1]
            all_models = self.storage.get('models')
            del all_models[to_drop]
            self.storage.set('models', all_models)

        else:
            raise Exception(f"Query type {type(statement)} not supported")

    def select_query(self, stmt) -> pd.DataFrame:
        _, _, target, model_url = self._get_model(stmt)

        if target not in [str(t) for t in stmt.targets]:
            raise Exception("Predictor will not be called, target column is not specified.")

        df = pd.DataFrame.from_dict({stmt.where.args[0].parts[0]: [stmt.where.args[1].value]})
        return self._call_model(df, model_url)


    def join(self, stmt, data_handler: BaseHandler, into: Optional[str] = None) -> pd.DataFrame:
        """
        Batch prediction using the output of a query passed to a data handler as input for the model.
        """  # noqa

        # tag data and predictive handlers
        if len(stmt.from_table.left.parts) == 1:
            model_clause = 'left'
            data_clause = 'right'
        else:
            model_clause = 'right'
            data_clause = 'left'
        model_alias = str(getattr(stmt.from_table, model_clause).alias)

        # get model input
        data_handler_table = getattr(stmt.from_table, data_clause).parts[-1]  # todo should be ".".join(...) if data handlers support more than one table
        data_handler_cols = list(set([t.parts[-1] for t in stmt.targets]))

        data_query = f"""SELECT {','.join(data_handler_cols)} FROM {data_handler_table}"""
        if stmt.where:
            data_query += f" WHERE {str(stmt.where)}"
        if stmt.limit:
            # todo integration should handle this depending on type of query... e.g. if it is TS, then we have to fetch correct groups first and limit later
            data_query += f" LIMIT {stmt.limit.value}"

        parsed_query = self.parser(data_query, dialect=self.dialect)
        model_input = pd.DataFrame.from_records(
            data_handler.select_query(
                parsed_query.targets,
                parsed_query.from_table,
                parsed_query.where
            ))

        # rename columns
        aliased_columns = list(model_input.columns)
        for col in stmt.targets:
            if str(col.parts[0]) != model_alias and col.alias is not None:
                # assumes mdb_sql will alert if there are two columns with the same alias
                aliased_columns[aliased_columns.index(col.parts[-1])] = str(col.alias)
        model_input.columns = aliased_columns

        # get model output
        _, _, _, model_url = self._get_model(stmt)
        predictions = self._call_model(model_input, model_url)

        # rename columns
        aliased_columns = list(predictions.columns)
        for col in stmt.targets:
            if col.parts[0] == model_alias and col.alias is not None:
                aliased_columns[aliased_columns.index('prediction')] = str(col.alias)
        predictions.columns = aliased_columns

        if into:
            try:
                data_handler.select_into(into, predictions)
            except Exception as e:
                print("Error when trying to store the JOIN output in data handler.")

        return predictions

    def _get_model(self, stmt):
        if type(stmt.from_table) == Join:
            model_name = stmt.from_table.right.parts[-1]
        else:
            model_name = stmt.from_table.parts[-1]

        mlflow_models = [model.name for model in self.connection.list_registered_models()]
        if not model_name in self.get_tables():
            raise Exception("Error, not found. Please create this predictor first.")
        elif not model_name in mlflow_models:
            raise Exception(
                "Cannot connect with the model, it might not served. Please serve it with MLflow and try again.")

        model = self.connection.get_registered_model(model_name)
        model_info = self.storage.get('models')[model_name]
        return model_name, model, model_info['target'], model_info['url']

    def _call_model(self, df, model_url):
        resp = requests.post(model_url,
                             data=df.to_json(orient='records'),
                             headers={'content-type': 'application/json; format=pandas-records'})
        answer: List[object] = resp.json()

        predictions = pd.DataFrame({'prediction': answer})
        out = df.join(predictions)
        return out



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'saslContinue': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        try:
            payload = query['payload'].decode()

            if len(payload) > 0:
                responce = session.scram.process_client_second_message(payload)
                responce = responce.encode()
                done = False
            else:
                responce = None
                done = True

            res = {
                'conversationId': 1,
                'done': done,
                'payload': responce,
                'ok': 1
            }
        except Exception as e:
            res = {
                'errmsg': str(e),
                'ok': 0
            }
        return res


responder = Responce()



from ..mysql_handler import Handler as MySQLHandler


class MariaDBHandler(MySQLHandler):
    """
    This handler handles connection and execution of the MariaDB statements.
    """
    name = 'mariadb'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



import mindsdb_sql
from mindsdb_sql.parser.ast import (
    Insert,
    Set,
    Alter,
    Update
)
from mindsdb_sql import parse_sql
from mindsdb_sql.planner import utils as planner_utils

from mindsdb.api.mysql.mysql_proxy.classes.sql_query import (
    Column,
    SQLQuery
)
from mindsdb.api.mysql.mysql_proxy.classes.sql_statement_parser import SqlStatementParser
from mindsdb.api.mysql.mysql_proxy.utilities import (
    ErBadDbError,
    SqlApiException,
    log
)


from mindsdb.api.mysql.mysql_proxy.executor.executor_commands import ExecuteCommands


class Executor:

    def __init__(self, session, sqlserver):
        self.session = session
        self.sqlserver = sqlserver

        self.query = None

        # returns
        self.columns = []
        self.params = []
        self.data = None
        self.state_track = None
        self.server_status = None

        self.is_executed = False

        # self.predictor_metadata = {}

        self.sql = ''
        self.sql_lower = ''

        self.command_executor = ExecuteCommands(self.session, self)

    def stmt_prepare(self, sql):

        resp = self.execute_external(sql)
        if resp is not None:
            # is already executed
            self.is_executed = True
            return

        self.parse(sql)

        # if not params
        params = planner_utils.get_query_params(self.query)
        if len(params) == 0:
            # execute immediately
            self.do_execute()

        else:
            # plan query
            # TODO less complex.
            #  planner is inside SQLQuery now.

            sqlquery = SQLQuery(
                self.query,
                session=self.session,
                execute=False
            )

            sqlquery.prepare_query()

            self.params = [
                Column(
                    alias=p.value,
                    type='str',
                    name=p.value,
                )
                for p in params
            ]

            # TODO:
            #   select * from mindsdb.predictors doesn't invoke prepare_steps and columns_list is empty
            self.columns = sqlquery.columns_list

    def stmt_execute(self, param_values):
        if self.is_executed:
            return

        # fill params
        self.query = planner_utils.fill_query_params(self.query, param_values)

        # execute query
        self.do_execute()

    def query_execute(self, sql):
        resp = self.execute_external(sql)
        if resp is not None:
            # is already executed
            self.is_executed = True
            return

        self.parse(sql)
        self.do_execute()

    def execute_external(self, sql):

        # not exec directly in integration
        return None

        # try exec in external integration
        if (
            isinstance(self.session.database, str)
            and len(self.session.database) > 0
            and self.session.database.lower() not in ('mindsdb',
                                                      'files',
                                                      'information_schema')
            and '@@' not in sql.lower()
            and (
                (
                    sql.lower().strip().startswith('select')
                    and 'from' in sql.lower()
                )
                or (
                    sql.lower().strip().startswith('show')
                    # and 'databases' in sql.lower()
                    and 'tables' in sql.lower()
                )
            )
        ):
            datanode = self.session.datahub.get(self.session.database)
            if datanode is None:
                raise ErBadDbError('Unknown database - %s' % self.session.database)

            # try parse or send raw sql
            try:
                sql = parse_sql(sql, dialect='mindsdb')
            except mindsdb_sql.exceptions.ParsingException:
                pass

            result, column_info = datanode.query(sql)
            columns = [
                Column(name=col['name'], type=col['type'])
                for col in column_info
            ]

            data = []
            if len(result) > 0:
                # columns = [{
                #     'table_name': '',
                #     'name': x,
                #     'type': TYPES.MYSQL_TYPE_VAR_STRING
                # } for x in result[0].keys()]

                data = [[str(value) for key, value in x.items()] for x in result]
            self.columns = columns
            self.data = data
            return True

    def parse(self, sql):
        self.sql = sql
        sql_lower = sql.lower()
        self.sql_lower = sql_lower.replace('`', '')

        try:
            try:
                self.query = parse_sql(sql, dialect='mindsdb')
            except Exception:
                self.query = parse_sql(sql, dialect='mysql')
        except Exception as e:
            # not all statements are parsed by parse_sql
            log.warning(f'SQL statement are not parsed by mindsdb_sql: {sql}')

            sql_list = [x for x in self.sql_lower.replace('\t', ' ').replace('\n', ' ').split(' ') if x not in ('', ' ')]
            if len(sql_list) > 1 and sql_list[0] == "show":
                raise SqlApiException(f"unknown command: {sql}")
            if len(sql_list) > 2 and " ".join(sql_list[:2]) == "create predictor":
                if 'predict' not in sql_list:
                    raise SqlApiException(f"'predict' field is mandatory: {sql}")
                # analyze predictor name
                if not sql_list[2][0].isalpha():
                    raise SqlApiException(f"predictor name must start from letter character: {sql}")

            # TODO
            if sql_lower == "set names 'utf8mb4' collate 'utf8mb4_general_ci'":
                self.query = Set()  # empty set command
            if sql_lower.startswith('alter table') and (
                    sql_lower.endswith('disable keys') or sql_lower.endswith('enable keys')):
                self.query = Set()  # empty set command

            st = SqlStatementParser(sql)
            keyword = st.keyword
            if keyword == 'set':
                self.query = Set()
            elif keyword == 'update':
                self.query = Update()
            elif keyword == 'insert':
                self.query = Insert(table='')
            elif keyword == 'alter':
                self.query = Alter()
            else:
                raise SqlApiException(f'SQL statement cannot be parsed by mindsdb_sql - {sql}: {e}') from e

            # not all statements are parse by parse_sql
            log.warning(f'SQL statement are not parsed by mindsdb_sql: {sql}')

            # == a place for workarounds ==
            # or run sql in integration without parsing

    def do_execute(self):
        # it can be already run at prepare state
        if self.is_executed:
            return

        ret = self.command_executor.execute_command(self.query)

        self.is_executed = True

        self.data = ret.data
        self.server_status = ret.status
        if ret.columns is not None:
            self.columns = ret.columns

        self.state_track = ret.state_track



from flask import request
from flask_restx import Resource

from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.handlers import ns_conf
from mindsdb.integrations.utilities.install import install_dependencies


@ns_conf.route('/')
class HandlersList(Resource):
    @ns_conf.doc('handlers_list')
    def get(self):
        '''List all db handlers'''
        handlers = request.integration_controller.get_handlers_import_status()
        result = []
        for handler_type, handler_meta in handlers.items():
            row = {'name': handler_type}
            row.update(handler_meta)
            result.append(row)
        return result


@ns_conf.route('/<handler_name>/install')
class InstallDependencies(Resource):
    @ns_conf.param('handler_name', 'Handler name')
    def post(self, handler_name):
        handler_import_status = self.request.integration_controller.get_handlers_import_status()
        if handler_name not in handler_import_status:
            return f'Unkown handler: {handler_name}', 400

        if handler_import_status[handler_name].get('import', {}).get('success', False) is True:
            return 'Installed', 200

        handler_meta = handler_import_status[handler_name]

        dependencies = handler_meta['import']['dependencies']
        if len(dependencies) == 0:
            return 'Installed', 200

        result = install_dependencies(dependencies)
        if result.get('success') is True:
            self.request.integration_controller._load_handler_modules()
            return 'Installed', 200
        return http_error(
            500,
            'Failed to install dependency',
            result.get('error_message', 'unknown error')
        )



import os
import shutil
import tempfile
import importlib
from pathlib import Path
from copy import deepcopy
import base64

from sqlalchemy import func

from mindsdb.interfaces.storage.db import session, Integration
from mindsdb.utilities.config import Config
from mindsdb.interfaces.storage.fs import FsStore
from mindsdb.utilities.fs import create_directory

from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.interfaces.database.views import ViewController
from mindsdb.utilities.with_kwargs_wrapper import WithKWArgsWrapper


class IntegrationController:
    @staticmethod
    def _is_not_empty_str(s):
        return isinstance(s, str) and len(s) > 0

    def __init__(self):
        self._load_handler_modules()

    def add(self, name, engine, data, company_id=None):
        bundle_path = data.get('secure_connect_bundle')
        if engine in ('cassandra', 'scylla') and self._is_not_empty_str(bundle_path):
            if os.path.isfile(bundle_path) is False:
                raise Exception(f'Can not get access to file: {bundle_path}')
            integrations_dir = Config()['paths']['integrations']

            p = Path(bundle_path)
            data['secure_connect_bundle'] = p.name

            integration_record = Integration(
                name=name,
                engine=engine,
                data=data,
                company_id=company_id
            )
            session.add(integration_record)
            session.commit()
            integration_id = integration_record.id

            folder_name = f'integration_files_{company_id}_{integration_id}'
            integration_dir = os.path.join(integrations_dir, folder_name)
            create_directory(integration_dir)
            shutil.copyfile(bundle_path, os.path.join(integration_dir, p.name))

            FsStore().put(
                folder_name,
                integration_dir,
                integrations_dir
            )
        elif engine in ('mysql', 'mariadb'):
            ssl = data.get('ssl')
            files = {}
            temp_dir = None
            if ssl is True:
                for key in ['ssl_ca', 'ssl_cert', 'ssl_key']:
                    if key not in data:
                        continue
                    if os.path.isfile(data[key]) is False:
                        if self._is_not_empty_str(data[key]) is False:
                            raise Exception("'ssl_ca', 'ssl_cert' and 'ssl_key' must be paths or inline certs")
                        if temp_dir is None:
                            temp_dir = tempfile.mkdtemp(prefix='integration_files_')
                        cert_file_name = data.get(f'{key}_name', f'{key}.pem')
                        cert_file_path = os.path.join(temp_dir, cert_file_name)
                        with open(cert_file_path, 'wt') as f:
                            f.write(data[key])
                        data[key] = cert_file_path
                    files[key] = data[key]
                    p = Path(data[key])
                    data[key] = p.name
            integration_record = Integration(
                name=name,
                engine=engine,
                data=data,
                company_id=company_id
            )
            session.add(integration_record)
            session.commit()
            integration_id = integration_record.id

            if len(files) > 0:
                integrations_dir = Config()['paths']['integrations']
                folder_name = f'integration_files_{company_id}_{integration_id}'
                integration_dir = os.path.join(integrations_dir, folder_name)
                create_directory(integration_dir)
                for file_path in files.values():
                    p = Path(file_path)
                    shutil.copyfile(file_path, os.path.join(integration_dir, p.name))
                FsStore().put(
                    folder_name,
                    integration_dir,
                    integrations_dir
                )
        else:
            integration_record = Integration(
                name=name,
                engine=engine,
                data=data,
                company_id=company_id
            )
            session.add(integration_record)
            session.commit()

    def modify(self, name, data, company_id):
        integration_record = session.query(Integration).filter_by(company_id=company_id, name=name).first()
        old_data = deepcopy(integration_record.data)
        for k in old_data:
            if k not in data:
                data[k] = old_data[k]

        integration_record.data = data
        session.commit()

    def delete(self, name, company_id=None):
        integration_record = session.query(Integration).filter_by(company_id=company_id, name=name).first()
        integrations_dir = Config()['paths']['integrations']
        folder_name = f'integration_files_{company_id}_{integration_record.id}'
        integration_dir = os.path.join(integrations_dir, folder_name)
        if os.path.isdir(integration_dir):
            shutil.rmtree(integration_dir)
        try:
            FsStore().delete(folder_name)
        except Exception:
            pass
        session.delete(integration_record)
        session.commit()

    def _get_integration_record_data(self, integration_record, sensitive_info=True):
        if integration_record is None or integration_record.data is None:
            return None
        data = deepcopy(integration_record.data)
        if data.get('password', None) is None:
            data['password'] = ''

        bundle_path = data.get('secure_connect_bundle')
        mysql_ssl_ca = data.get('ssl_ca')
        mysql_ssl_cert = data.get('ssl_cert')
        mysql_ssl_key = data.get('ssl_key')
        if (
            data.get('type') in ('mysql', 'mariadb')
            and (
                self._is_not_empty_str(mysql_ssl_ca)
                or self._is_not_empty_str(mysql_ssl_cert)
                or self._is_not_empty_str(mysql_ssl_key)
            )
            or data.get('type') in ('cassandra', 'scylla')
            and bundle_path is not None
        ):
            fs_store = FsStore()
            integrations_dir = Config()['paths']['integrations']
            folder_name = f'integration_files_{integration_record.company_id}_{integration_record.id}'
            integration_dir = os.path.join(integrations_dir, folder_name)
            fs_store.get(
                folder_name,
                integration_dir,
                integrations_dir
            )

        if not sensitive_info:
            if 'password' in data:
                data['password'] = None
            if (
                data.get('type') == 'redis'
                and isinstance(data.get('connection'), dict)
                and 'password' in data['connection']
            ):
                data['connection'] = None

        return {
            'id': integration_record.id,
            'name': integration_record.name,
            'engine': integration_record.engine,
            'date_last_update': deepcopy(integration_record.updated_at),
            'connection_data': data
        }

    def get(self, name, company_id=None, sensitive_info=True, case_sensitive=False):
        if case_sensitive:
            integration_record = session.query(Integration).filter_by(company_id=company_id, name=name).first()
        else:
            integration_record = session.query(Integration).filter(
                (Integration.company_id == company_id)
                & (func.lower(Integration.name) == func.lower(name))
            ).first()
        return self._get_integration_record_data(integration_record, sensitive_info)

    def get_all(self, company_id=None, sensitive_info=True):
        integration_records = session.query(Integration).filter_by(company_id=company_id).all()
        integration_dict = {}
        for record in integration_records:
            if record is None or record.data is None:
                continue
            integration_dict[record.name] = self._get_integration_record_data(record, sensitive_info)
        return integration_dict

    def check_connections(self):
        connections = {}
        for integration_name, integration_meta in self.get_all().items():
            handler = self.create_handler(
                handler_type=integration_meta['engine'],
                connection_data=integration_meta['connection_data']
            )
            status = handler.check_connection()
            connections[integration_name] = status.get('success', False)
        return connections

    def create_handler(self, name: str = None, handler_type: str = None,
                       connection_data: dict = {}, company_id: int = None):
        fs_store = FsStore()

        handler_ars = dict(
            name=name,
            db_store=None,
            fs_store=fs_store,
            connection_data=connection_data
        )

        if handler_type == 'files':
            handler_ars['file_controller'] = WithKWArgsWrapper(
                FileController(),
                company_id=company_id
            )
        elif handler_type == 'views':
            handler_ars['view_controller'] = WithKWArgsWrapper(
                ViewController(),
                company_id=company_id
            )

        return self.handler_modules[handler_type].Handler(**handler_ars)

    def get_handler(self, name, company_id=None, case_sensitive=False):
        if case_sensitive:
            integration_record = session.query(Integration).filter_by(company_id=company_id, name=name).first()
        else:
            integration_record = session.query(Integration).filter(
                (Integration.company_id == company_id)
                & (func.lower(Integration.name) == func.lower(name))
            ).first()
        if integration_record is None:
            raise Exception(f'Unknown integration: {name}')
        integration_meta = self._get_integration_record_data(integration_record, True)

        integration_engine = integration_meta['engine']
        integration_name = integration_meta['name']

        if integration_engine not in self.handler_modules:
            raise Exception(f"Cant find handler for '{integration_name}' ({integration_engine})")

        handler = self.create_handler(
            name=integration_name,
            handler_type=integration_engine,
            connection_data=integration_meta['connection_data'],
            company_id=company_id
        )

        return handler

    def _load_handler_modules(self):
        mindsdb_path = Path(importlib.util.find_spec('mindsdb').origin).parent
        handlers_path = mindsdb_path.joinpath('integrations/handlers')
        self.handler_modules = {}
        self.handlers_import_status = {}
        for hanlder_dir in handlers_path.iterdir():
            if hanlder_dir.is_dir() is False or hanlder_dir.name.startswith('__'):
                continue
            handler_folder_name = str(hanlder_dir.name)
            handler_name = handler_folder_name
            if handler_name.endswith('_handler'):
                handler_name = handler_name[:-8]
            dependencies = []
            requirements_txt = hanlder_dir.joinpath('requirements.txt')
            if requirements_txt.is_file():
                with open(str(requirements_txt), 'rt') as f:
                    dependencies = [x.strip(' \t\n') for x in f.readlines()]
                    dependencies = [x for x in dependencies if len(x) > 0]
            try:
                handler_module = importlib.import_module(f'mindsdb.integrations.handlers.{handler_folder_name}')
                self.handler_modules[handler_module.name] = handler_module
                import_error = None
                if hasattr(handler_module, 'import_error'):
                    import_error = handler_module.import_error
                handler_meta = {
                    'import': {
                        'success': import_error is None,
                        'folder': handler_folder_name,
                        'dependencies': dependencies
                    },
                    'version': handler_module.version
                }
                if import_error is not None:
                    handler_meta['import']['error_message'] = str(import_error)

                for attr in ('connection_args_example', 'connection_args', 'description', 'name', 'type', 'title'):
                    if hasattr(handler_module, attr):
                        handler_meta[attr] = getattr(handler_module, attr)
                if 'name' not in handler_meta:
                    handler_meta['name'] = handler_name

                # region icon
                if hasattr(handler_module, 'icon_path'):
                    icon_path = hanlder_dir.joinpath(handler_module.icon_path)
                    handler_meta['icon'] = {
                        'name': icon_path.name,
                        'type': icon_path.name[icon_path.name.rfind('.') + 1:].lower()
                    }
                    if handler_meta['icon']['type'] == 'svg':
                        with open(str(icon_path), 'rt') as f:
                            handler_meta['icon']['data'] = f.read()
                    else:
                        with open(str(icon_path), 'rb') as f:
                            handler_meta['icon']['data'] = base64.b64encode(f.read()).decode('utf-8')
                # endregion
            except Exception as e:
                handler_meta = {
                    'import': {
                        'success': False,
                        'error_message': str(e),
                        'folder': handler_folder_name,
                        'dependencies': dependencies
                    },
                    'name': handler_name
                }

            if handler_meta.get('name') in ('files', 'views', 'lightwood'):
                handler_meta['permanent'] = True
            else:
                handler_meta['permanent'] = False

            self.handlers_import_status[handler_meta['name']] = handler_meta

    def get_handlers_import_status(self):
        return self.handlers_import_status



class ModelInterface():
    def __init__(self):
        from mindsdb.interfaces.model.model_controller import ModelController
        self.controller = ModelController(False)

    def create(self, *args, **kwargs):
        return self.controller.create(*args, **kwargs)

    def learn(self, *args, **kwargs):
        return self.controller.learn(*args, **kwargs)

    def adjust(self, *args, **kwargs):
        return self.controller.adjust(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self.controller.predict(*args, **kwargs)

    def analyse_dataset(self, *args, **kwargs):
        return self.controller.analyse_dataset(*args, **kwargs)

    def get_model_data(self, *args, **kwargs):
        return self.controller.get_model_data(*args, **kwargs)

    def get_model_description(self, *args, **kwargs):
        return self.controller.get_model_description(*args, **kwargs)

    def get_models(self, *args, **kwargs):
        return self.controller.get_models(*args, **kwargs)

    def delete_model(self, *args, **kwargs):
        return self.controller.delete_model(*args, **kwargs)

    def rename_model(self, *args, **kwargs):
        return self.controller.rename_model(*args, **kwargs)

    def update_model(self, *args, **kwargs):
        return self.controller.update_model(*args, **kwargs)

    def generate_predictor(self, *args, **kwargs):
        return self.controller.generate_predictor(*args, **kwargs)

    def edit_json_ai(self, *args, **kwargs):
        return self.controller.edit_json_ai(*args, **kwargs)

    def edit_code(self, *args, **kwargs):
        return self.controller.edit_code(*args, **kwargs)

    def fit_predictor(self, *args, **kwargs):
        return self.controller.fit_predictor(*args, **kwargs)

    def code_from_json_ai(self, *args, **kwargs):
        return self.controller.code_from_json_ai(*args, **kwargs)

    def export_predictor(self, *args, **kwargs):
        return self.controller.export_predictor(*args, **kwargs)

    def import_predictor(self, *args, **kwargs):
        return self.controller.import_predictor(*args, **kwargs)


ray_based = False

'''
Notes: Remove ray from actors are getting stuck
try:
    from mindsdb_worker.cluster.ray_interface import ModelInterfaceRay
    import ray
    try:
        ray.init(ignore_reinit_error=True, address='auto')
    except Exception:
        ray.init(ignore_reinit_error=True)
    ModelInterface = ModelInterfaceRay
    ray_based = True
except Exception as e:
    ModelInterface = ModelInterfaceNativeImport
    ray_based = False
'''



def get_column_in_case(columns, name):
    '''
    '''
    candidates = []
    name_lower = name.lower()
    for column in columns:
        if column.lower() == name_lower:
            candidates.append(column)
    if len(candidates) != 1:
        return None
    return candidates[0]



import os
import json
from copy import deepcopy

from mindsdb.utilities.fs import create_directory


def _merge_key_recursive(target_dict, source_dict, key):
    if key not in target_dict:
        target_dict[key] = source_dict[key]
    elif not isinstance(target_dict[key], dict) or not isinstance(source_dict[key], dict):
        target_dict[key] = source_dict[key]
    else:
        for k in list(source_dict[key].keys()):
            _merge_key_recursive(target_dict[key], source_dict[key], k)


def _merge_configs(original_config, override_config):
    original_config = deepcopy(original_config)
    for key in list(override_config.keys()):
        _merge_key_recursive(original_config, override_config, key)
    return original_config


class Config():
    def __init__(self):
        self.config_path = os.environ['MINDSDB_CONFIG_PATH']
        if self.config_path == 'absent':
            self._override_config = {}
        else:
            with open(self.config_path, 'r') as fp:
                self._override_config = json.load(fp)

        self._default_config = {
            'permanent_storage': {
                'location': 'local'
            },
            'paths': {},
            "log": {
                "level": {
                    "console": "INFO",
                    "file": "DEBUG",
                    "db": "WARNING"
                }

            },
            "debug": False,
            "integrations": {},
            "api": {
                "http": {
                    "host": "127.0.0.1",
                    "port": "47334"
                },
                "mysql": {
                    "host": "127.0.0.1",
                    "password": "",
                    "port": "47335",
                    "user": "mindsdb",
                    "database": "mindsdb",
                    "ssl": True
                },
                "mongodb": {
                    "host": "127.0.0.1",
                    "port": "47336",
                    "database": "mindsdb"
                }
            },
            "cache": {
                "type": "local"
            },
            "force_dataset_removing": False
        }

        self._default_config['paths']['root'] = os.environ['MINDSDB_STORAGE_DIR']
        self._default_config['paths']['storage'] = os.path.join(self._default_config['paths']['root'], 'storage')
        self._default_config['paths']['datasources'] = os.path.join(self._default_config['paths']['root'], 'datasources')
        self._default_config['paths']['predictors'] = os.path.join(self._default_config['paths']['root'], 'predictors')
        self._default_config['paths']['static'] = os.path.join(self._default_config['paths']['root'], 'static')
        self._default_config['paths']['tmp'] = os.path.join(self._default_config['paths']['root'], 'tmp')
        self._default_config['paths']['log'] = os.path.join(self._default_config['paths']['root'], 'log')
        self._default_config['paths']['storage_dir'] = self._default_config['paths']['root']
        self._default_config['paths']['cache'] = os.path.join(self._default_config['paths']['root'], 'cache')
        self._default_config['paths']['integrations'] = os.path.join(self._default_config['paths']['root'], 'integrations')
        self._default_config['storage_dir'] = self._default_config['paths']['root']

        for path_name in self._default_config['paths']:
            create_directory(self._default_config['paths'][path_name])

        self._config = _merge_configs(self._default_config, self._override_config)

    def __getitem__(self, key):
        return self._config[key]

    def get(self, key, default=None):
        return self._config.get(key, default)

    def get_all(self):
        return self._config

    @property
    def paths(self):
        return self._config['paths']



import unittest

from lightwood.mixer import LightGBM

from mindsdb.utilities.config import Config
from mindsdb.integrations.handlers.lightwood_handler.lightwood_handler.lightwood_handler import LightwoodHandler
from mindsdb.integrations.handlers.lightwood_handler.lightwood_handler.utils import load_predictor
from mindsdb.integrations.handlers.mysql_handler.mysql_handler import MySQLHandler


# TODO: bring all train+predict queries in mindsdb_sql test suite
# TODO: drop all models and tables when closing tests
class LightwoodHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.handler = LightwoodHandler('test_ludwig_handler')
        cls.config = Config()

        kwargs = {"connection_data": {
            "host": "localhost",
            "port": "3306",
            "user": "root",
            "password": "root",
            "database": "test",
            "ssl": False
        }}
        cls.sql_handler_name = 'test_handler'
        cls.data_handler = MySQLHandler(cls.sql_handler_name, **kwargs)

        cls.target_1 = 'rental_price'
        cls.data_table_name_1 = 'home_rentals_subset'
        cls.test_model_name_1 = 'test_lightwood_home_rentals'
        cls.test_model_name_1b = 'test_lightwood_home_rentals_custom'

        cls.target_2 = 'Traffic'
        cls.data_table_name_2 = 'arrival'
        cls.test_model_name_2 = 'test_lightwood_arrivals'
        cls.model_2_into_table = 'test_join_tsmodel_into_lw'

        # todo: connect to global state instead
        cls.MDB_CURRENT_HANDLERS = {
            cls.sql_handler_name: MySQLHandler(cls.sql_handler_name, **{"connection_data": {
                "host": "localhost",
                "port": "3306",
                "user": "root",
                "password": "root",
                "database": "test",
                "ssl": False
            }})
        }
        cls.data_handler = cls.MDB_CURRENT_HANDLERS[cls.sql_handler_name]

    def connect_handler(self):
        config = Config()
        self.handler.connect(config={'path': config['paths']['root'], 'name': 'lightwood_handler.db'})

    def test_0_connect(self):
        self.connect_handler()

    def test_1_drop_predictor(self):
        model_name = 'lw_test_predictor'
        try:
            print('dropping predictor...')
            self.handler.native_query(f"DROP PREDICTOR {model_name}")
        except:
            print('failed to drop')
            pass

    def test_21_train_predictor(self):
        if self.test_model_name_1 not in self.handler.get_tables():
            query = f"CREATE PREDICTOR {self.test_model_name_1} FROM {self.sql_handler_name} (SELECT * FROM test.{self.data_table_name_1}) PREDICT {self.target_1}"
            self.handler.native_query(query)

    def test_22_retrain_predictor(self):
        query = f"RETRAIN {self.test_model_name_1}"
        self.handler.native_query(query)

    def test_31_list_tables(self):
        print(self.handler.get_tables())

    def test_32_describe_table(self):
        print(self.handler.describe_table(f'{self.test_model_name_1}'))

    def test_41_query_predictor_single_where_condition(self):
        query = f"SELECT target from {self.test_model_name_1} WHERE sqft=100"
        parsed = self.handler.parser(query, dialect=self.handler.dialect)
        predicted = self.handler.query(parsed)['data_frame']

    def test_42_query_predictor_multi_where_condition(self):
        query = f"SELECT target from {self.test_model_name_1} WHERE sqft=100 AND number_of_rooms=2 AND number_of_bathrooms=1"
        parsed = self.handler.parser(query, dialect=self.handler.dialect)
        predicted = self.handler.query(parsed)['data_frame']

    def test_51_join_predictor_into_table(self):
        into_table = 'test_join_into_lw'
        query = f"SELECT tb.{self.target_1} as predicted, ta.{self.target_1} as truth, ta.sqft from {self.sql_handler_name}.{self.data_table_name_1} AS ta JOIN {self.test_model_name_1} AS tb LIMIT 10"
        parsed = self.handler.parser(query, dialect=self.handler.dialect)
        predicted = self.handler.join(parsed, self.data_handler, into=into_table)

        # checks whether `into` kwarg does insert into the table or not
        q = f"SELECT * FROM {into_table}"
        qp = self.handler.parser(q, dialect='mysql')
        assert len(self.data_handler.query(qp).data_frame) > 0

    def test_23_train_predictor_custom_jsonai(self):
        if self.test_model_name_1b not in self.handler.get_tables():
            using_str = 'model.args={"submodels": [{"module": "LightGBM", "args": {"stop_after": 12, "fit_on_dev": True}}]}'
            query = f'CREATE PREDICTOR {self.test_model_name_1b} FROM {self.sql_handler_name} (SELECT * FROM test.{self.data_table_name_1}) PREDICT {self.target_1} USING {using_str}'
            self.handler.native_query(query)

        m = load_predictor(self.handler.storage.get('models')[self.test_model_name_1b], self.test_model_name_1b)
        assert len(m.ensemble.mixers) == 1
        assert isinstance(m.ensemble.mixers[0], LightGBM)

    def test_24_train_ts_predictor(self):
        target = 'Traffic'
        oby = 'T'
        gby = 'Country'
        window = 8
        horizon = 4

        if self.test_model_name_2 not in self.handler.get_tables():
            query = f'CREATE PREDICTOR {self.test_model_name_2} FROM {self.sql_handler_name} (SELECT * FROM test.{self.data_table_name_2}) PREDICT {target} ORDER BY {oby} GROUP BY {gby} WINDOW {window} HORIZON {horizon}'
            self.handler.native_query(query)

        p = self.handler.storage.get('models')
        m = load_predictor(p[self.test_model_name_2], self.test_model_name_2)
        assert m.problem_definition.timeseries_settings.is_timeseries

    def test_52_join_predictor_ts_into(self):
        self.connect_handler()
        target = 'Traffic'
        oby = 'T'
        query = f"SELECT tb.{target} as predicted, ta.{target} as truth, ta.{oby} from {self.sql_handler_name}.{self.data_table_name_2} AS ta JOIN mindsdb.{self.test_model_name_2} AS tb ON 1=1 WHERE ta.{oby} > LATEST LIMIT 10"
        parsed = self.handler.parser(query, dialect=self.handler.dialect)
        predicted = self.handler.join(parsed, self.data_handler, into=self.model_2_into_table)




"""views

Revision ID: 47f97b83cee4
Revises: 17c3d2384711
Create Date: 2022-01-26 12:07:05.075977

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db



# revision identifiers, used by Alembic.
revision = '47f97b83cee4'
down_revision = '17c3d2384711'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('view',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('company_id', sa.Integer(), nullable=True),
    sa.Column('query', sa.String(), nullable=False),
    sa.Column('datasource_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['datasource_id'], ['datasource.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('view')
    # ### end Alembic commands ###



from mindsdb.interfaces.stream.redis.redisdb import Redis
from mindsdb.interfaces.stream.kafka.kafkadb import Kafka
from mindsdb.utilities.log import log as logger
from mindsdb.utilities.config import Config
from mindsdb.interfaces.database.integrations import IntegrationController
from mindsdb.utilities.with_kwargs_wrapper import WithKWArgsWrapper


class StreamController():
    known_dbs = {
        'redis': Redis,
        'kafka': Kafka
    }

    def __init__(self, company_id):
        self.config = Config()
        self.company_id = company_id
        self.integration_controller = WithKWArgsWrapper(
            IntegrationController(), company_id=company_id
        )

    def setup(self, db_alias):
        try:
            integration_meta = self.integration_controller.get(db_alias)
            if integration_meta is None:
                raise Exception(f'Unkonw database integration: {db_alias}')
            if integration_meta.get('engine') not in self.known_dbs:
                raise Exception(f'Unkonw database integration type for: {db_alias}')
            self.known_dbs[integration_meta['engine']](self.config, db_alias, integration_meta).setup()
        except Exception as e:
            logger.warning('Failed to setup stream for ' + db_alias + f', error: {e}')



class Responder():
    def __init__(self, when=None, result=None):
        if when is not None:
            self.when = when
        if result is not None:
            self.result = result
        if not hasattr(self, 'when') or (not isinstance(self.when, dict) and not callable(self.when)):
            raise ValueError("Responder attr 'when' must be dict or function.")
        if not hasattr(self, 'result') or (not isinstance(self.result, dict) and not callable(self.result)):
            raise ValueError("Responder attr 'result' must be dict or function.")

    def match(self, query):
        """ check, if this 'responder' can be used to answer or current request

        query (dict): request document

        return bool
        """
        if isinstance(self.when, dict):
            for key, value in self.when.items():
                if key not in query:
                    return False
                if callable(value):
                    if not value(query[key]):
                        return False
                elif value != query[key]:
                    return False
            return True
        else:
            return self.when(query)

    def handle(self, query, args, env, session):
        """ making answer based on params:

        query (dict): document(s) from request
        args (dict): all other significant information from request: flags, collection name, rows to return, etc
        env (dict): config, model_interface instance, and other mindsdb related stuff
        session (object): current session

        returns documents as dict or list of dicts
        """
        if isinstance(self.result, dict):
            return self.result
        else:
            return self.result(query, args, env, session)



from distutils.version import LooseVersion
import requests
import os
import shutil
import threading
import webbrowser
from zipfile import ZipFile
from pathlib import Path
import traceback
import tempfile
# import concurrent.futures
from flask import Flask, url_for, make_response
from flask.json import dumps
from flask_restx import Api

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.interfaces.model.model_interface import ModelInterface
from mindsdb.interfaces.database.integrations import IntegrationController
from mindsdb.interfaces.file.file_controller import FileController
from mindsdb.utilities.ps import is_pid_listen_port, wait_func_is_true
from mindsdb.utilities.telemetry import inject_telemetry_to_static
from mindsdb.utilities.config import Config
from mindsdb.utilities.log import get_log
from mindsdb.interfaces.storage.db import session
from mindsdb.utilities.json_encoder import CustomJSONEncoder


class Swagger_Api(Api):
    """
    This is a modification of the base Flask Restplus Api class due to the issue described here
    https://github.com/noirbizarre/flask-restplus/issues/223
    """
    @property
    def specs_url(self):
        return url_for(self.endpoint("specs"), _external=False)


def custom_output_json(data, code, headers=None):
    resp = make_response(dumps(data), code)
    resp.headers.extend(headers or {})
    return resp


def get_last_compatible_gui_version() -> LooseVersion:
    log = get_log('http')

    try:
        res = requests.get('https://mindsdb-web-builds.s3.amazonaws.com/compatible-config.json', timeout=5)
    except (ConnectionError, requests.exceptions.ConnectionError) as e:
        print(f'Is no connection. {e}')
        return False
    except Exception as e:
        print(f'Is something wrong with getting compatible-config.json: {e}')
        return False

    if res.status_code != 200:
        print(f'Cant get compatible-config.json: returned status code = {res.status_code}')
        return False

    try:
        versions = res.json()
    except Exception as e:
        print(f'Cant decode compatible-config.json: {e}')
        return False

    current_mindsdb_lv = LooseVersion(mindsdb_version)

    try:
        gui_versions = {}
        max_mindsdb_lv = None
        max_gui_lv = None
        for el in versions['mindsdb']:
            if el['mindsdb_version'] is None:
                gui_lv = LooseVersion(el['gui_version'])
            else:
                mindsdb_lv = LooseVersion(el['mindsdb_version'])
                gui_lv = LooseVersion(el['gui_version'])
                if mindsdb_lv.vstring not in gui_versions or gui_lv > gui_versions[mindsdb_lv.vstring]:
                    gui_versions[mindsdb_lv.vstring] = gui_lv
                if max_mindsdb_lv is None or max_mindsdb_lv < mindsdb_lv:
                    max_mindsdb_lv = mindsdb_lv
            if max_gui_lv is None or max_gui_lv < gui_lv:
                max_gui_lv = gui_lv

        all_mindsdb_lv = [LooseVersion(x) for x in gui_versions.keys()]
        all_mindsdb_lv.sort()

        if current_mindsdb_lv.vstring in gui_versions:
            gui_version_lv = gui_versions[current_mindsdb_lv.vstring]
        elif current_mindsdb_lv > all_mindsdb_lv[-1]:
            gui_version_lv = max_gui_lv
        else:
            lower_versions = {key: value for key, value in gui_versions.items() if LooseVersion(key) < current_mindsdb_lv}
            if len(lower_versions) == 0:
                gui_version_lv = gui_versions[all_mindsdb_lv[0].vstring]
            else:
                all_lower_versions = [LooseVersion(x) for x in lower_versions.keys()]
                gui_version_lv = gui_versions[all_lower_versions[-1].vstring]
    except Exception as e:
        log.error(f'Error in compatible-config.json structure: {e}')
        return False
    return gui_version_lv


def get_current_gui_version() -> LooseVersion:
    config = Config()
    static_path = Path(config['paths']['static'])
    version_txt_path = static_path.joinpath('version.txt')

    current_gui_version = None
    if version_txt_path.is_file():
        with open(version_txt_path, 'rt') as f:
            current_gui_version = f.readline()

    current_gui_lv = None if current_gui_version is None else LooseVersion(current_gui_version)

    return current_gui_lv


def download_gui(destignation, version):
    if isinstance(destignation, str):
        destignation = Path(destignation)
    log = get_log('http')
    dist_zip_path = str(destignation.joinpath('dist.zip'))
    bucket = "https://mindsdb-web-builds.s3.amazonaws.com/"

    resources = [{
        'url': bucket + 'dist-V' + version + '.zip',
        'path': dist_zip_path
    }]

    def get_resources(resource):
        response = requests.get(resource['url'])
        if response.status_code != requests.status_codes.codes.ok:
            raise Exception(f"Error {response.status_code} GET {resource['url']}")
        open(resource['path'], 'wb').write(response.content)
    try:
        for r in resources:
            get_resources(r)
    except Exception as e:
        log.error(f'Error during downloading files from s3: {e}')
        return False

    static_folder = destignation
    static_folder.mkdir(mode=0o777, exist_ok=True, parents=True)
    ZipFile(dist_zip_path).extractall(static_folder)

    if static_folder.joinpath('dist').is_dir():
        shutil.move(str(destignation.joinpath('dist').joinpath('index.html')), static_folder)
        shutil.move(str(destignation.joinpath('dist').joinpath('assets')), static_folder)
        shutil.rmtree(destignation.joinpath('dist'))

    os.remove(dist_zip_path)

    version_txt_path = destignation.joinpath('version.txt')  # os.path.join(destignation, 'version.txt')
    with open(version_txt_path, 'wt') as f:
        f.write(version)

    return True

    '''
    # to make downloading faster download each resource in a separate thread
    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        future_to_url = {executor.submit(get_resources, r): r for r in resources}
        for future in concurrent.futures.as_completed(future_to_url):
            res = future.result()
            if res is not None:
                raise res
    '''


def initialize_static():
    success = update_static()
    session.close()
    return success


def update_static():
    ''' Update Scout files basing on compatible-config.json content.
        Files will be downloaded and updated if new version of GUI > current.
        Current GUI version stored in static/version.txt.
    '''
    config = Config()
    log = get_log('http')
    static_path = Path(config['paths']['static'])

    last_gui_version_lv = get_last_compatible_gui_version()
    current_gui_version_lv = get_current_gui_version()

    if last_gui_version_lv is False:
        return False

    if current_gui_version_lv is not None:
        if current_gui_version_lv >= last_gui_version_lv:
            return True

    log.info(f'New version of GUI available ({last_gui_version_lv.vstring}). Downloading...')

    temp_dir = tempfile.mkdtemp(prefix='mindsdb_gui_files_')
    success = download_gui(temp_dir, last_gui_version_lv.vstring)
    if success is False:
        shutil.rmtree(temp_dir)
        return False

    temp_dir_for_rm = tempfile.mkdtemp(prefix='mindsdb_gui_files_')
    shutil.rmtree(temp_dir_for_rm)
    shutil.copytree(str(static_path), temp_dir_for_rm)
    shutil.rmtree(str(static_path))
    shutil.copytree(temp_dir, str(static_path))
    shutil.rmtree(temp_dir_for_rm)

    log.info(f'GUI version updated to {last_gui_version_lv.vstring}')
    return True


def initialize_flask(config, init_static_thread, no_studio):
    # Apparently there's a bug that causes the static path not to work if it's '/' -- https://github.com/pallets/flask/issues/3134, I think '' should achieve the same thing (???)
    if no_studio:
        app = Flask(
            __name__
        )
    else:
        static_path = os.path.join(config['paths']['static'], 'static/')
        if os.path.isabs(static_path) is False:
            static_path = os.path.join(os.getcwd(), static_path)
        app = Flask(
            __name__,
            static_url_path='/static',
            static_folder=static_path
        )

    app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 60
    app.config['SWAGGER_HOST'] = 'http://localhost:8000/mindsdb'
    app.json_encoder = CustomJSONEncoder

    authorizations = {
        'apikey': {
            'type': 'session',
            'in': 'query',
            'name': 'session'
        }
    }

    api = Swagger_Api(
        app,
        authorizations=authorizations,
        security=['apikey'],
        url_prefix=':8000',
        prefix='/api',
        doc='/doc/'
    )

    api.representations['application/json'] = custom_output_json

    port = config['api']['http']['port']
    host = config['api']['http']['host']

    # NOTE rewrite it, that hotfix to see GUI link
    if not no_studio:
        log = get_log('http')
        if host in ('', '0.0.0.0'):
            url = f'http://127.0.0.1:{port}/'
        else:
            url = f'http://{host}:{port}/'
        log.info(f' - GUI available at {url}')

        pid = os.getpid()
        x = threading.Thread(target=_open_webbrowser, args=(url, pid, port, init_static_thread, config['paths']['static']), daemon=True)
        x.start()

    return app, api


def initialize_interfaces(app):
    app.original_model_interface = ModelInterface()
    app.original_integration_controller = IntegrationController()
    app.original_file_controller = FileController()
    config = Config()
    app.config_obj = config


def _open_webbrowser(url: str, pid: int, port: int, init_static_thread, static_folder):
    """Open webbrowser with url when http service is started.

    If some error then do nothing.
    """
    init_static_thread.join()
    inject_telemetry_to_static(static_folder)
    logger = get_log('http')
    try:
        is_http_active = wait_func_is_true(func=is_pid_listen_port, timeout=10,
                                           pid=pid, port=port)
        if is_http_active:
            webbrowser.open(url)
    except Exception as e:
        logger.error(f'Failed to open {url} in webbrowser with exception {e}')
        logger.error(traceback.format_exc())
    session.close()



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'isMaster': helpers.is_true}

    result = {
        "ismaster": True,
        "minWireVersion": 0,
        "maxWireVersion": 9,
        "ok": 1
    }


responder = Responce()



import json

from flask import Response


def http_error(status_code, title, detail=''):
    ''' Wrapper for error responce acoording with RFC 7807 (https://tools.ietf.org/html/rfc7807)

        :param status_code: int - http status code for response
        :param title: str
        :param detail: str

        :return: flask Response object
    '''
    return Response(
        response=json.dumps({
            'title': title,
            'detail': detail
        }),
        status=status_code,
        headers={
            'Content-Type': 'application/problem+json'
        }
    )



def empty_fn(*args, **kwargs):
    pass


try:
    from mindsdb.utilities.hooks.after_predict import after_predict
except ImportError:
    after_predict = empty_fn


try:
    from mindsdb.utilities.hooks.after_api_query import after_api_query
except ImportError:
    after_api_query = empty_fn



import unittest
from mindsdb.integrations.handlers.questdb_handler.questdb_handler import QuestDBHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class QuestDBHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "127.0.0.1",
            "port": "8812",
            "user": "admin",
            "password": "quest",
            "database": "questdb"
        }
        cls.handler = QuestDBHandler('test_questdb_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()
    
    def test_1_describe_table(self):
        described = self.handler.describe_table("house_rentals_data")
        assert described['type'] is not RESPONSE_TYPE.ERROR
    
    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables['type'] is not RESPONSE_TYPE.ERROR

    def test_3_select_query(self):
        query = "SELECT * FROM house_rentals_data WHERE 'id'='1'"
        result = self.handler.query(query)
        assert len(result) > 0




from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController

from mindsdb.api.mysql.mysql_proxy.executor.executor_commands import ExecuteCommands

def run_sql_command(mindsdb_env, ast_query):
    # empty object
    server_obj = type('', (), {})()

    server_obj.original_integration_controller = mindsdb_env['origin_integration_controller']
    server_obj.original_model_interface = mindsdb_env['origin_model_interface']
    server_obj.original_view_controller = mindsdb_env['origin_view_controller']

    sql_session = SessionController(
        server=server_obj,
        company_id=mindsdb_env['company_id']
    )
    sql_session.database = 'mindsdb'

    command_executor = ExecuteCommands(sql_session, executor=None)
    ret = command_executor.execute_command(ast_query)
    if ret.error_code is not None:
        raise Exception(ret.error_message)

    column_names = [
        c.name is c.alias is None or c.alias
        for c in ret.columns
    ]

    data = []
    for row in ret.data:
        data.append(dict(zip(column_names, row)))

    return data




import gc
import sys
import dill

from lightwood.api.high_level import _module_from_code


def unpack_jsonai_old_args(json_ai_override):
    while '.' in str(list(json_ai_override.keys())):
        for k in list(json_ai_override.keys()):
            if '.' in k:
                nks = k.split('.')
                obj = json_ai_override
                for nk in nks[:-1]:
                    if nk not in obj:
                        obj[nk] = {}
                    obj = obj[nk]
                obj[nks[-1]] = json_ai_override[k]
                del json_ai_override[k]


def load_predictor(predictor_dict, name):
    try:
        module_name = None
        return dill.loads(predictor_dict['predictor'])
    except Exception as e:
        module_name = str(e).lstrip("No module named '").split("'")[0]

        try:
            del sys.modules[module_name]
        except Exception:
            pass

        gc.collect()
        _module_from_code(predictor_dict['code'], module_name)
        return dill.loads(predictor_dict['predictor'])



import pandas as pd
from mindsdb_sql.parser.ast import Identifier, Constant, Star, Select, Join


def recur_get_conditionals(args: list, values):
    """ Gets all the specified data from an arbitrary amount of AND clauses inside the WHERE statement """  # noqa
    if isinstance(args[0], Identifier) and isinstance(args[1], Constant):
        values[args[0].parts[0]] = [args[1].value]
    else:
        for op in args:
            values = {**values, **recur_get_conditionals([*op.args], {})}
    return values


def get_aliased_columns(aliased_columns, model_alias, targets, mode=None):
    """ This method assumes mdb_sql will alert if there are two columns with the same alias """  # noqa
    for col in targets:
        if mode == 'input':
            if str(col.parts[0]) != model_alias and col.alias is not None:
                aliased_columns[aliased_columns.index(col.parts[-1])] = str(col.alias)

        if mode == 'output':
            if str(col.parts[0]) == model_alias and col.alias is not None:
                aliased_columns[aliased_columns.index('prediction')] = str(col.alias)

    return aliased_columns


def default_data_gather(handler, query):
    """ Default method to gather data from another handler into a pandas dataframe """  # noqa
    records = handler.query(query).data_frame
    df = pd.DataFrame.from_records(records)
    return df


def get_join_input(query, model, model_aliases, data_handler, data_side):
    target_cols = set()
    for t in query.targets:
        if t.parts[0] not in model_aliases:
            if t.parts[-1] == Star():
                target_cols = [Star()]
                break
            else:
                target_cols.add(t.parts[-1])

    if target_cols != [Star()]:
        target_cols = [Identifier(col) for col in target_cols]

    data_handler_table = getattr(query.from_table, data_side).parts[-1]
    data_query = Select(
        targets=target_cols,
        from_table=Identifier(data_handler_table),
        where=query.where,
        group_by=query.group_by,
        having=query.having,
        order_by=query.order_by,
        offset=query.offset,
        limit=query.limit
    )

    model_input = pd.DataFrame.from_records(
        data_handler.query(data_query).data_frame
    )

    return model_input


def get_model_name(handler, stmt):
     """ Discern between joined entities to retrieve model name, alias and the clause side it is on. """
     side = None
     models = handler.get_tables() # .data_frame['model_name'].values
     if type(stmt.from_table) == Join:
         model_name = stmt.from_table.right.parts[-1]
         side = 'right'
         if model_name not in models:
             model_name = stmt.from_table.left.parts[-1]
             side = 'left'
         alias = str(getattr(stmt.from_table, side).alias)
     else:
         model_name = stmt.from_table.parts[-1]
         alias = None  # todo: fix this

     if model_name not in models:
         raise Exception("Error, not found. Please create this predictor first.")

     return model_name, alias, side


from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'listIndexes': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        return {
            'cursor': [{
                'v': 2,
                'key': {
                    '_id': 1
                },
                'name': '_id_',
                'ns': f"{query['$db']}.{query['listIndexes']}"
            }],
            'ok': 1,
        }


responder = Responce()



import os
import json
import datetime

import numpy as np
from sqlalchemy import create_engine, orm, types, UniqueConstraint
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, DateTime, Boolean, Index
from sqlalchemy.sql.expression import null
from sqlalchemy.sql.schema import ForeignKey


if os.environ['MINDSDB_DB_CON'].startswith('sqlite:'):
    engine = create_engine(os.environ['MINDSDB_DB_CON'], echo=False)
else:
    engine = create_engine(os.environ['MINDSDB_DB_CON'], convert_unicode=True, pool_size=30, max_overflow=200, echo=False)
Base = declarative_base()
session = scoped_session(sessionmaker(bind=engine, autoflush=True))
Base.query = session.query_property()


# Source: https://stackoverflow.com/questions/26646362/numpy-array-is-not-json-serializable
class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class Array(types.TypeDecorator):
    ''' Float Type that replaces commas with  dots on input '''
    impl = types.String

    def process_bind_param(self, value, dialect):  # insert
        if isinstance(value, str):
            return value
        elif value is None:
            return value
        else:
            return ',|,|,'.join(value)

    def process_result_value(self, value, dialect):  # select
        return value.split(',|,|,') if value is not None else None


class Json(types.TypeDecorator):
    ''' Float Type that replaces commas with  dots on input '''
    impl = types.String

    def process_bind_param(self, value, dialect):  # insert
        return json.dumps(value, cls=NumpyEncoder) if value is not None else None

    def process_result_value(self, value, dialect):  # select
        return json.loads(value) if value is not None else None


class Semaphor(Base):
    __tablename__ = 'semaphor'

    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    entity_type = Column('entity_type', String)
    entity_id = Column('entity_id', Integer)
    action = Column(String)
    company_id = Column(Integer)
    uniq_const = UniqueConstraint('entity_type', 'entity_id')


class Predictor(Base):
    __tablename__ = 'predictor'

    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    name = Column(String)
    data = Column(Json)  # A JSON -- should be everything returned by `get_model_data`, I think
    to_predict = Column(Array)
    company_id = Column(Integer)
    mindsdb_version = Column(String)
    native_version = Column(String)
    integration_id = Column(ForeignKey('integration.id', name='fk_integration_id'), nullable=True)
    fetch_data_query = Column(String)
    is_custom = Column(Boolean)
    learn_args = Column(Json)
    update_status = Column(String, default='up_to_date')

    json_ai = Column(Json, nullable=True)
    code = Column(String, nullable=True)
    lightwood_version = Column(String, nullable=True)
    dtype_dict = Column(Json, nullable=True)
    uniq_const = UniqueConstraint('name', 'company_id', name='unique_name_company_id')


class Log(Base):
    __tablename__ = 'log'

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, default=datetime.datetime.now)
    log_type = Column(String)  # log, info, warning, traceback etc
    source = Column(String)  # file + line
    company_id = Column(Integer)
    payload = Column(String)
    created_at_index = Index("some_index", "created_at_index")


class Integration(Base):
    __tablename__ = 'integration'
    id = Column(Integer, primary_key=True)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    name = Column(String, nullable=False)
    engine = Column(String, nullable=False)
    data = Column(Json)
    company_id = Column(Integer)
    uniq_const = UniqueConstraint('name', 'company_id', name='unique_name_company_id')


class Stream(Base):
    __tablename__ = 'stream'
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    stream_in = Column(String, nullable=False)
    stream_out = Column(String, nullable=False)
    anomaly_stream = Column(String)
    integration = Column(String)
    predictor = Column(String, nullable=False)
    company_id = Column(Integer)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    created_at = Column(DateTime, default=datetime.datetime.now)
    type = Column(String, default='unknown')
    connection_info = Column(Json, default={})
    learning_params = Column(Json, default={})
    learning_threshold = Column(Integer, default=0)


class File(Base):
    __tablename__ = 'file'
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    company_id = Column(Integer)
    source_file_path = Column(String, nullable=False)
    file_path = Column(String, nullable=False)
    row_count = Column(Integer, nullable=False)
    columns = Column(Json, nullable=False)
    created_at = Column(DateTime, default=datetime.datetime.now)
    updated_at = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    uniq_const = UniqueConstraint('name', 'company_id', name='unique_name_company_id')


class View(Base):
    __tablename__ = 'view'
    id = Column(Integer, primary_key=True)
    name = Column(String, nullable=False)
    company_id = Column(Integer)
    query = Column(String, nullable=False)
    uniq_const = UniqueConstraint('name', 'company_id', name='unique_name_company_id')


# DDL is changing through migrations
# Base.metadata.create_all(engine)
# orm.configure_mappers()



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .view_handler import ViewHandler as Handler
from .__about__ import __version__ as version, __description__ as description


title = 'View'
name = 'views'
type = HANDLER_TYPE.DATA

__all__ = [
    'Handler', 'version', 'name', 'type', 'title',
    'description', 'connection_args_example'
]



from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import ERR


# base exception for unknown error
class SqlApiUnknownError(Exception):
    err_code = ERR.ER_UNKNOWN_ERROR


# base exception for known error
class SqlApiException(Exception):
    err_code = ERR.ER_SYNTAX_ERROR


class ErBadDbError(SqlApiException):
    err_code = ERR.ER_BAD_DB_ERROR

class ErBadTableError(SqlApiException):
    err_code = ERR.ER_BAD_DB_ERROR

class ErKeyColumnDoesNotExist(SqlApiException):
    err_code = ERR.ER_KEY_COLUMN_DOES_NOT_EXIST

class ErTableExistError(SqlApiException):
    err_code = ERR.ER_TABLE_EXISTS_ERROR

class ErDubFieldName(SqlApiException):
    err_code = ERR.ER_DUP_FIELDNAME

class ErDbDropDelete(SqlApiException):
    err_code = ERR.ER_DB_DROP_DELETE

class ErNonInsertableTable(SqlApiException):
    err_code = ERR.ER_NON_INSERTABLE_TABLE

class ErNotSupportedYet(SqlApiException):
    err_code = ERR.ER_NOT_SUPPORTED_YET

class ErSqlSyntaxError(SqlApiException):
    err_code = ERR.ER_SYNTAX_ERROR

class ErSqlWrongArguments(SqlApiException):
    err_code = ERR.ER_WRONG_ARGUMENTS

class ErWrongCharset(SqlApiException):
    err_code = ERR.ER_UNKNOWN_CHARACTER_SET

class ErLogicError(SqlApiException):
    err_code = ERR.ER_WRONG_USAGE



from functools import partial

import pandas as pd

from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser.ast import BinaryOperation, Select, Identifier, Constant

from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import get_all_tables
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.mindsdb_datanode import MindsDBDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.integration_datanode import IntegrationDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import TablesRow, TABLES_ROW_TYPE
from mindsdb.api.mysql.mysql_proxy.utilities import exceptions as exc


class InformationSchemaDataNode(DataNode):
    type = 'INFORMATION_SCHEMA'

    information_schema = {
        'SCHEMATA': ['CATALOG_NAME', 'SCHEMA_NAME', 'DEFAULT_CHARACTER_SET_NAME', 'DEFAULT_COLLATION_NAME', 'SQL_PATH'],
        'TABLES': ['TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'TABLE_TYPE', 'ENGINE', 'VERSION', 'ROW_FORMAT', 'TABLE_ROWS', 'AVG_ROW_LENGTH', 'DATA_LENGTH', 'MAX_DATA_LENGTH', 'INDEX_LENGTH', 'DATA_FREE', 'AUTO_INCREMENT', 'CREATE_TIME', 'UPDATE_TIME', 'CHECK_TIME', 'TABLE_COLLATION', 'CHECKSUM', 'CREATE_OPTIONS', 'TABLE_COMMENT'],
        'COLUMNS': ['TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'COLUMN_NAME', 'ORDINAL_POSITION', 'COLUMN_DEFAULT', 'IS_NULLABLE', 'DATA_TYPE', 'CHARACTER_MAXIMUM_LENGTH', 'CHARACTER_OCTET_LENGTH', 'NUMERIC_PRECISION', 'NUMERIC_SCALE', 'DATETIME_PRECISION', 'CHARACTER_SET_NAME', 'COLLATION_NAME', 'COLUMN_TYPE', 'COLUMN_KEY', 'EXTRA', 'PRIVILEGES', 'COLUMN_COMMENT', 'GENERATION_EXPRESSION'],
        'EVENTS': ['EVENT_CATALOG', 'EVENT_SCHEMA', 'EVENT_NAME', 'DEFINER', 'TIME_ZONE', 'EVENT_BODY', 'EVENT_DEFINITION', 'EVENT_TYPE', 'EXECUTE_AT', 'INTERVAL_VALUE', 'INTERVAL_FIELD', 'SQL_MODE', 'STARTS', 'ENDS', 'STATUS', 'ON_COMPLETION', 'CREATED', 'LAST_ALTERED', 'LAST_EXECUTED', 'EVENT_COMMENT', 'ORIGINATOR', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'ROUTINES': ['SPECIFIC_NAME', 'ROUTINE_CATALOG', 'ROUTINE_SCHEMA', 'ROUTINE_NAME', 'ROUTINE_TYPE', 'DATA_TYPE', 'CHARACTER_MAXIMUM_LENGTH', 'CHARACTER_OCTET_LENGTH', 'NUMERIC_PRECISION', 'NUMERIC_SCALE', 'DATETIME_PRECISION', 'CHARACTER_SET_NAME', 'COLLATION_NAME', 'DTD_IDENTIFIER', 'ROUTINE_BODY', 'ROUTINE_DEFINITION', 'EXTERNAL_NAME', 'EXTERNAL_LANGUAGE', 'PARAMETER_STYLE', 'IS_DETERMINISTIC', 'SQL_DATA_ACCESS', 'SQL_PATH', 'SECURITY_TYPE', 'CREATED', 'LAST_ALTERED', 'SQL_MODE', 'ROUTINE_COMMENT', 'DEFINER', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'TRIGGERS': ['TRIGGER_CATALOG', 'TRIGGER_SCHEMA', 'TRIGGER_NAME', 'EVENT_MANIPULATION', 'EVENT_OBJECT_CATALOG', 'EVENT_OBJECT_SCHEMA', 'EVENT_OBJECT_TABLE', 'ACTION_ORDER', 'ACTION_CONDITION', 'ACTION_STATEMENT', 'ACTION_ORIENTATION', 'ACTION_TIMING', 'ACTION_REFERENCE_OLD_TABLE', 'ACTION_REFERENCE_NEW_TABLE', 'ACTION_REFERENCE_OLD_ROW', 'ACTION_REFERENCE_NEW_ROW', 'CREATED', 'SQL_MODE', 'DEFINER', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'PLUGINS': ['PLUGIN_NAME', 'PLUGIN_VERSION', 'PLUGIN_STATUS', 'PLUGIN_TYPE', 'PLUGIN_TYPE_VERSION', 'PLUGIN_LIBRARY', 'PLUGIN_LIBRARY_VERSION', 'PLUGIN_AUTHOR', 'PLUGIN_DESCRIPTION', 'PLUGIN_LICENSE', 'LOAD_OPTION', 'PLUGIN_MATURITY', 'PLUGIN_AUTH_VERSION'],
        'ENGINES': ['ENGINE', 'SUPPORT', 'COMMENT', 'TRANSACTIONS', 'XA', 'SAVEPOINTS'],
        'KEY_COLUMN_USAGE': ['CONSTRAINT_CATALOG', 'CONSTRAINT_SCHEMA', 'CONSTRAINT_NAME', 'TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'COLUMN_NAME', 'ORDINAL_POSITION', 'POSITION_IN_UNIQUE_CONSTRAINT', 'REFERENCED_TABLE_SCHEMA', 'REFERENCED_TABLE_NAME', 'REFERENCED_COLUMN_NAME'],
        'STATISTICS': ['TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'NON_UNIQUE', 'INDEX_SCHEMA', 'INDEX_NAME', 'SEQ_IN_INDEX', 'COLUMN_NAME', 'COLLATION', 'CARDINALITY', 'SUB_PART', 'PACKED', 'NULLABLE', 'INDEX_TYPE', 'COMMENT', 'INDEX_COMMENT', 'IS_VISIBLE', 'EXPRESSION'],
        'CHARACTER_SETS': ['CHARACTER_SET_NAME', 'DEFAULT_COLLATE_NAME', 'DESCRIPTION', 'MAXLEN'],
        'COLLATIONS': ['COLLATION_NAME', 'CHARACTER_SET_NAME', 'ID', 'IS_DEFAULT', 'IS_COMPILED', 'SORTLEN', 'PAD_ATTRIBUTE'],
    }

    def __init__(self, session):
        self.session = session
        self.integration_controller = session.integration_controller
        self.view_interface = session.view_interface
        self.persis_datanodes = {
            'mindsdb': MindsDBDataNode(
                session.model_interface,
                session.integration_controller
            ),
            'files': IntegrationDataNode(
                'files',
                ds_type='file',
                integration_controller=self.session.integration_controller
            ),
            'views': IntegrationDataNode(
                'views',
                ds_type='view',
                integration_controller=self.session.integration_controller
            )
        }

        self.get_dataframe_funcs = {
            'TABLES': self._get_tables,
            'COLUMNS': self._get_columns,
            'SCHEMATA': self._get_schemata,
            'ENGINES': self._get_engines,
            'CHARACTER_SETS': self._get_charsets,
            'COLLATIONS': self._get_collations,
        }
        for table_name in self.information_schema:
            if table_name not in self.get_dataframe_funcs:
                self.get_dataframe_funcs[table_name] = partial(self._get_empty_table, table_name)

    def __getitem__(self, key):
        return self.get(key)

    def get(self, name):
        name_lower = name.lower()

        if name.lower() == 'information_schema':
            return self

        if name_lower in self.persis_datanodes:
            return self.persis_datanodes[name_lower]

        integration_names = self.integration_controller.get_all().keys()
        for integration_name in integration_names:
            if integration_name.lower() == name_lower:
                datasource = self.integration_controller.get(name=integration_name)
                return IntegrationDataNode(
                    integration_name,
                    ds_type=datasource['engine'],
                    integration_controller=self.session.integration_controller
                )

        return None

    def has_table(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return True
        return False

    def get_table_columns(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return self.information_schema[tn]
        raise exc.ErTableExistError(f'Table information_schema.{tableName} does not exists')

    def get_integrations_names(self):
        integration_names = self.integration_controller.get_all().keys()
        # remove files and views from list to prevent doubling in 'select from INFORMATION_SCHEMA.TABLES'
        return [
            x.lower()
            for x in integration_names
            if x not in ('files', 'views')
        ]

    def _get_tables(self, query: ASTNode = None):
        columns = self.information_schema['TABLES']

        target_table = None
        if type(query) == Select and type(query.where) == BinaryOperation and query.where.op == 'and':
            for arg in query.where.args:
                if (
                    type(arg) == BinaryOperation and arg.op == '='
                    and type(arg.args[0]) == Identifier
                    and arg.args[0].parts[0].upper() == 'TABLES'
                    and arg.args[0].parts[1].upper() == 'TABLE_SCHEMA'
                    and type(arg.args[1]) == Constant
                ):
                    target_table = arg.args[1].value
                    break

        data = []
        for name in self.information_schema.keys():
            if target_table is not None and target_table != name:
                continue
            row = TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.SYSTEM_VIEW, TABLE_NAME=name)
            data.append(row.to_list())

        for ds_name, ds in self.persis_datanodes.items():
            if target_table is not None and target_table != ds_name:
                continue
            ds_tables = ds.get_tables()
            if len(ds_tables) == 0:
                continue
            elif isinstance(ds_tables[0], dict):
                ds_tables = [TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x['name']) for x in ds_tables]
            elif isinstance(ds_tables, list) and len(ds_tables) > 0 and isinstance(ds_tables[0], str):
                ds_tables = [TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x) for x in ds_tables]
            for row in ds_tables:
                row.TABLE_SCHEMA = ds_name
                data.append(row.to_list())

        for ds_name in self.get_integrations_names():
            if target_table is not None and target_table != ds_name:
                continue
            try:
                ds = self.get(ds_name)
                ds_tables = ds.get_tables()
                for row in ds_tables:
                    row.TABLE_SCHEMA = ds_name
                    data.append(row.to_list())
            except Exception:
                print(f"Can't get tables from '{ds_name}'")

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_columns(self, query: ASTNode = None):
        columns = self.information_schema['COLUMNS']

        # NOTE there is a lot of types in mysql, but listed below should be enough for our purposes
        row_templates = {
            'text': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'varchar', 1024, 3072, None, None, None, 'utf8', 'utf8_bin', 'varchar(1024)', None, None, 'select', None, None],
            'timestamp': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', 'CURRENT_TIMESTAMP', 'YES', 'timestamp', None, None, None, None, 0, None, None, 'timestamp', None, None, 'select', None, None],
            'bigint': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'bigint', None, None, 20, 0, None, None, None, 'bigint unsigned', None, None, 'select', None, None],
            'float': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'float', None, None, 12, 0, None, None, None, 'float', None, None, 'select', None, None]
        }

        result = []

        for table_name in self.information_schema:
            table_columns = self.information_schema[table_name]
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'information_schema'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        mindsb_dn = self.get('MINDSDB')
        for table_name in mindsb_dn.get_tables():
            table_columns = mindsb_dn.get_table_columns(table_name)
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'mindsdb'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        mindsb_dn = self.get('FILES')
        for table_name in mindsb_dn.get_tables():
            table_columns = mindsb_dn.get_table_columns(table_name)
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'files'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        df = pd.DataFrame(result, columns=columns)
        return df

    def _get_schemata(self, query: ASTNode = None):
        columns = self.information_schema['SCHEMATA']
        data = [
            ['def', 'information_schema', 'utf8', 'utf8_general_ci', None]
        ]

        # permanent databases
        data.append(['def', 'mindsdb', 'utf8mb4', 'utf8mb4_0900_ai_ci', None])

        integration_names = self.integration_controller.get_all().keys()
        for database_name in integration_names:
            data.append(['def', database_name, 'utf8mb4', 'utf8mb4_0900_ai_ci', None])

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_engines(self, query: ASTNode = None):
        columns = self.information_schema['ENGINES']
        data = [['InnoDB', 'DEFAULT', 'Supports transactions, row-level locking, and foreign keys', 'YES', 'YES', 'YES']]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_charsets(self, query: ASTNode = None):
        columns = self.information_schema['CHARACTER_SETS']
        data = [
            ['utf8', 'UTF-8 Unicode', 'utf8_general_ci', 3],
            ['latin1', 'cp1252 West European', 'latin1_swedish_ci', 1],
            ['utf8mb4', 'UTF-8 Unicode', 'utf8mb4_general_ci', 4]
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_collations(self, query: ASTNode = None):
        columns = self.information_schema['COLLATIONS']
        data = [
            ['utf8_general_ci', 'utf8', 33, 'Yes', 'Yes', 1, 'PAD SPACE'],
            ['latin1_swedish_ci', 'latin1', 8, 'Yes', 'Yes', 1, 'PAD SPACE']
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_empty_table(self, table_name, query: ASTNode = None):
        columns = self.information_schema[table_name]
        data = []

        df = pd.DataFrame(data, columns=columns)
        return df

    def query(self, query: ASTNode):
        query_tables = get_all_tables(query)

        if len(query_tables) != 1:
            raise exc.ErBadTableError(f'Only one table can be used in query to information_schema: {query}')

        table_name = query_tables[0].upper()

        if table_name not in self.get_dataframe_funcs:
            raise exc.ErNotSupportedYet('Information schema: Not implemented.')

        dataframe = self.get_dataframe_funcs[table_name](query=query)

        try:
            data = query_df(dataframe, query, session=self.session)
        except Exception as e:
            print(f'Exception! {e}')
            return [], []

        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in data.dtypes.items()
        ]

        return data.to_dict(orient='records'), columns_info



import json
from dateutil.parser import parse as parse_datetime
from flask import request
from flask_restx import Resource, abort
from mindsdb.utilities.log import log
from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.predictors import ns_conf
from mindsdb.api.http.namespaces.entitites.predictor_metadata import (
    predictor_query_params,
    put_predictor_params,
    put_predictor_metadata
)
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


@ns_conf.route('/')
class PredictorList(Resource):
    @ns_conf.doc('list_predictors')
    def get(self):
        '''List all predictors'''
        models = request.model_interface.get_models()
        return models


@ns_conf.route('/<name>')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class Predictor(Resource):
    @ns_conf.doc('get_predictor')
    def get(self, name):
        model = request.model_interface.get_model_data(name)

        for k in ['train_end_at', 'updated_at', 'created_at']:
            if k in model and model[k] is not None:
                model[k] = parse_datetime(model[k])

        return model

    @ns_conf.doc('delete_predictor')
    def delete(self, name):
        '''Remove predictor'''
        request.model_interface.delete_model(name)
        return '', 200

    @ns_conf.doc('put_predictor')
    @ns_conf.expect(put_predictor_metadata)
    def put(self, name):
        '''Learning new predictor'''
        data = request.json
        to_predict = data.get('to_predict')

        try:
            kwargs = data.get('kwargs')
        except Exception:
            kwargs = None

        if isinstance(kwargs, dict) is False:
            kwargs = {}

        if 'equal_accuracy_for_all_output_categories' not in kwargs:
            kwargs['equal_accuracy_for_all_output_categories'] = True

        if 'advanced_args' not in kwargs:
            kwargs['advanced_args'] = {}

        if 'use_selfaware_model' not in kwargs['advanced_args']:
            kwargs['advanced_args']['use_selfaware_model'] = False

        integration_name = data.get('integration')
        query = data.get('query')
        if isinstance(integration_name, str) is False or isinstance(query, str) is False:
            return http_error(400, 'Error', 'Parameters should contain integration and query')

        integration_meta = request.integration_controller.get(integration_name)
        if integration_meta is None:
            return http_error(400, 'Error', f"Cant get integration '{integration_name}'")
        handler = request.integration_controller.get_handler(integration_name)
        if handler is None:
            return http_error(400, 'Error', f"Cant get integration '{integration_name}'")

        result = handler.native_query(query)

        if result.type != RESPONSE_TYPE.TABLE:
            raise Exception(f'Error during query: {result.get("error_message")}')

        df = result.data_frame

        request.model_interface.learn(
            name, df, to_predict, integration_id=integration_meta['id'],
            fetch_data_query=query, kwargs=kwargs, user_class=request.user_class
        )

        return '', 200


@ns_conf.route('/<name>/update')
@ns_conf.param('name', 'Update predictor')
class PredictorUpdate(Resource):
    @ns_conf.doc('Update predictor')
    def get(self, name):
        msg = request.model_interface.update_model(name)
        return {
            'message': msg
        }


@ns_conf.route('/<name>/adjust')
@ns_conf.param('name', 'The predictor identifier')
class PredictorAdjust(Resource):
    @ns_conf.doc('post_predictor_adjust', params=predictor_query_params)
    def post(self, name):
        return abort(410, 'Method is not available')
        # data = request.json

        # ds_name = data.get('data_source_name') if data.get('data_source_name') is not None else data.get('from_data')
        # from_data = request.default_store.get_datasource_obj(ds_name, raw=True)

        # if from_data is None:
        #     return {'message': f'Can not find datasource: {ds_name}'}, 400

        # model_names = [x['name'] for x in request.model_interface.get_models()]
        # if name not in model_names:
        #     return abort(404, f'Predictor "{name}" doesn\'t exist',)

        # request.model_interface.adjust(
        #     name,
        #     from_data,
        #     request.default_store.get_datasource(ds_name)['id']
        # )

        # return '', 200


@ns_conf.route('/<name>/predict')
@ns_conf.param('name', 'The predictor identifier')
class PredictorPredict(Resource):
    @ns_conf.doc('post_predictor_predict', params=predictor_query_params)
    def post(self, name):
        '''Queries predictor'''
        when = request.json.get('when')

        # list is a required type for TS prediction
        if isinstance(when, (dict, list)) is False or len(when) == 0:
            return 'No data provided for the predictions', 400

        results = request.model_interface.predict(name, when, 'explain')
        return results


@ns_conf.route('/<name>/predict_datasource')
@ns_conf.param('name', 'The predictor identifier')
class PredictorPredictFromDataSource(Resource):
    @ns_conf.doc('post_predictor_predict', params=predictor_query_params)
    def post(self, name):
        return abort(410, 'Method is not available')
        # data = request.json
        # use_raw = False

        # from_data = request.default_store.get_datasource_obj(data.get('data_source_name'), raw=use_raw)
        # if from_data is None:
        #     abort(400, 'No valid datasource given')

        # results = request.model_interface.predict(name, from_data, 'explain')
        # return results


@ns_conf.route('/<name>/rename')
@ns_conf.param('name', 'The predictor identifier')
class PredictorDownload(Resource):
    @ns_conf.doc('get_predictor_download')
    def get(self, name):
        try:
            new_name = request.args.get('new_name')
            request.model_interface.rename_model(name, new_name)
        except Exception as e:
            return str(e), 400

        return f'Renamed model to {new_name}', 200


@ns_conf.route('/generate/<name>')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorGenerate(Resource):
    def put(self, name):
        return abort(410, 'Method is not available')
        # problem_definition = request.json['problem_definition']
        # datasource_name = request.json['data_source_name']

        # from_data = request.default_store.get_datasource_obj(
        #     datasource_name,
        #     raw=True
        # )
        # datasource = request.default_store.get_datasource(datasource_name)

        # request.model_interface.generate_predictor(
        #     name,
        #     from_data,
        #     datasource['id'],
        #     problem_definition,
        #     request.json.get('join_learn_process', False)
        # )

        # return '', 200


@ns_conf.route('/<name>/edit/json_ai')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorEditJsonAI(Resource):
    def put(self, name):
        request.model_interface.edit_json_ai(name, request.json['json_ai'])
        return '', 200


@ns_conf.route('/<name>/edit/code')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorEditCode(Resource):
    def put(self, name):
        request.model_interface.edit_code(name, request.json['code'])
        return '', 200


@ns_conf.route('/<name>/train')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorTrain(Resource):
    def put(self, name):
        return abort(410, 'Method is not available')

        # for param in ['data_source_name']:
        #     if param not in request.json:
        #         return abort(400, 'Please provide {}'.format(param))

        # from_data = request.default_store.get_datasource_obj(
        #     request.json['data_source_name'],
        #     raw=True
        # )

        # request.model_interface.fit_predictor(name, from_data, request.json.get('join_learn_process', False))
        # return '', 200


@ns_conf.route('/<name>/export')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorExport(Resource):
    def get(self, name):
        payload: json = request.model_interface.export_predictor(name)
        return payload, 200


@ns_conf.route('/<name>/import')
@ns_conf.param('name', 'The predictor identifier')
@ns_conf.response(404, 'predictor not found')
class PredictorImport(Resource):
    def put(self, name):
        request.model_interface.import_predictor(name, request.json)
        return '', 200



import unittest

from mindsdb.utilities.config import Config
from mindsdb.integrations.handlers.ludwig_handler.ludwig_handler import LudwigHandler
from mindsdb.integrations.handlers.mysql_handler.mysql_handler import MySQLHandler


class LudwigHandlerTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.handler = LudwigHandler('test_ludwig_handler')
        cls.config = Config()

        kwargs = {"connection_data": {
            "host": "localhost",
            "port": "3306",
            "user": "root",
            "password": "root",
            "database": "test",
            "ssl": False
        }}
        cls.sql_handler_name = 'test_handler'
        cls.target = 'rental_price'
        cls.data_table_name = 'home_rentals_subset'
        cls.data_handler = MySQLHandler(cls.sql_handler_name, **kwargs)
        cls.test_model_name = 'test_ludwig_home_rentals'

    def test_1_connect(self):
        self.assertTrue(
            self.handler.connect(config={
                    'path': self.config['paths']['root'],
                    'name': 'test_name'
                }).success
        )

    def test_2_drop_predictor(self):
        try:
            self.handler.native_query(f"DROP PREDICTOR {self.test_model_name}")
        except:
            target = 'rental_price'
            query = f"CREATE PREDICTOR {self.test_model_name} FROM {self.sql_handler_name} (SELECT * FROM test.{self.data_table_name}) PREDICT {self.target}"  # noqa
            self.handler.native_query(query)
            self.test_2_drop_predictor()

    def test_3_create_predictor(self):
        if self.test_model_name in self.handler.get_tables().data_frame['model_name']:
            print('dropping...')
            self.handler.native_query(f"DROP PREDICTOR {self.test_model_name}")
        query = f"CREATE PREDICTOR {self.test_model_name} FROM {self.sql_handler_name} (SELECT * FROM test.{self.data_table_name}) PREDICT {self.target}"  # noqa
        self.handler.native_query(query)

    def test_4_get_tables(self):
        self.handler.get_tables()

    def test_5_get_columns(self):
        self.handler.get_columns(f'{self.test_model_name}')

    def test_6_select_from(self):
        self.assertTrue(self.data_handler.check_connection())
        # ludwig needs input for all columns in the original training df
        query = f"SELECT rental_price from {self.test_model_name} WHERE sqft=200 AND number_of_bathrooms=2 AND number_of_rooms=2 AND neighborhood='westbrae' AND location='great' AND days_on_market=10"
        parsed = self.handler.parser(query, dialect=self.handler.dialect)
        predicted = self.handler.query(parsed)

    def test_7_join(self):
        self.assertTrue(self.data_handler.check_connection())

        into_table = 'test_ludwig_join_into'
        query = f"SELECT tb.rental_price as predicted, ta.* from {self.sql_handler_name}.{self.data_table_name} AS ta JOIN mindsdb.{self.test_model_name} as tb LIMIT 10"

        parsed = self.handler.parser(query, dialect=self.handler.dialect)
        predicted = self.handler.join(parsed, self.data_handler, into=into_table)

        # TODO: restore
        # q = f"SELECT * FROM {into_table}"
        # qp = self.handler.parser(q, dialect='mysql')
        # assert len(self.data_handler.query(qp)) > 0
        #
        # try:
        #     self.data_handler.native_query(f"DROP TABLE test.{into_table}")
        # except:
        #     pass




from mindsdb_sql.parser.ast import *

from .mongodb_query import MongoQuery


class MongodbRender:

    def to_mongo_query(self, node):
        if isinstance(node, Select):
            return self.select(node)
        raise NotImplementedError(f'Unknown statement: {node.__name__}')

    def select(self, node):
        # collection
        if not isinstance(node.from_table, Identifier):
            raise NotImplementedError(f'Not supported from {node.from_table}')

        collection = node.from_table.parts[-1]

        # filter
        filters = {}

        if node.where is not None:
            filters = self.handle_where(node.where)

        group = {}
        project = {'_id': 0}
        if node.distinct:
            # is group by distinct fields
            group = {'_id': {}}
            # project = {'_id': 0} #  hide _id

        if node.targets is not None:
            for col in node.targets:
                if isinstance(col, Star):
                    # show all
                    project = {}
                    break
                if isinstance(col, Identifier):
                    name = col.parts[-1]
                    if col.alias is None:
                        alias = name
                    else:
                        alias = col.alias.parts[-1]

                    project[alias] = f'${name}'  # project field

                    # group by distinct fields
                    if node.distinct:
                        group['_id'][name] = f'${name}'  # group field
                        group[name] = {'$first': f'${name}'}  # show field

                elif isinstance(col, Constant):
                    val = str(col.value)  # str because int is interpreted as index
                    if col.alias is None:
                        alias = val
                    else:
                        alias = col.alias.parts[-1]
                    project[alias] = val


        if node.group_by is not None:
            # TODO
            raise NotImplementedError(f'Group {node.group_by}')

        sort = {}
        if node.order_by is not None:
            for col in node.order_by:
                name = col.field.parts[-1]
                direction = 1 if col.direction.upper() == 'ASC' else -1
                sort[name] = direction

        # compose mongo query

        mquery = MongoQuery(collection)

        method = 'aggregate'
        arg = []
        if filters:
            arg.append({"$match": filters})

        if group:
            arg.append({"$group": group})

        if project:
            arg.append({"$project": project})

        if sort:
            arg.append({"$sort": sort})

        if node.offset is not None:
            arg.append({"$skip": int(node.offset.value)})

        if node.limit is not None:
            arg.append({"$limit": int(node.limit.value)})

        mquery.add_step({
            'method': method,
            'args': [arg]
        })

        return mquery

    def handle_where(self, node):
        # todo UnaryOperation, function
        if not type(node) in [BinaryOperation]:
            raise NotImplementedError(f'Not supported type {type(node)}')

        # logic operation
        op = node.op.lower()
        arg1, arg2 = node.args

        if op in ('and', 'or'):
            query1 = self.handle_where(arg1)
            query2 = self.handle_where(arg2)

            ops = {
                'and': '$and',
                'or': '$or',
            }
            query = {ops[op]: [query1, query2]}
            return query

        ops_map = {
            '>=': '$gte',
            '>': '$gt',
            '<': '$lt',
            '<=': '$lte',
            '<>': '$ne',
            '!=': '$ne',
            '=': '$eq',
            '==': '$eq',
            'is': '$eq',
            'is not': '$ne',
        }

        if isinstance(arg1, Identifier):
            var_name = arg1.parts[-1]
            # is simple operation
            if isinstance(arg2, Constant):
                # identifier and constant

                val = arg2.value
                if op in ('=', '=='):
                    pass
                elif op in ops_map:
                    op2 = ops_map[op]
                    val = {op2: val}
                else:
                    raise NotImplementedError(f'Not supported operator {op}')

                return {var_name: val}

            # is IN condition
            elif isinstance(arg2, Tuple):
                # it should be IN, NOT IN
                ops = {
                    'in': '$in',
                    'not in': '$nin'
                }
                # must be list of Constants
                values = [
                    i.value
                    for i in arg2.items
                ]

                if op in ops:
                    op2 = ops[op]
                    cond = {op2: values}
                else:
                    raise NotImplementedError(f'Not supported operator {op}')

                return {var_name: cond}

        # try to make expression

        val1 = self.where_element_convert(arg1)
        val2 = self.where_element_convert(arg2)

        if op in ops_map:
            op2 = ops_map[op]
        else:
            raise NotImplementedError(f'Not supported operator {op}')

        return {
            '$expr': {
                op2: [val1, val2]
            }
        }

    def where_element_convert(self, node):
        if isinstance(node, Identifier):
            return f'${node.parts[-1]}'
        elif isinstance(node, Latest):
            return f'LATEST'
        elif isinstance(node, Constant):
            return node.value
        else:
            raise NotImplementedError(f'Unknown where element {node}')




from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'count': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        collection = query['count']

        count = 0
        if collection == 'predictors':
            count = len(mindsdb_env['model_interface'].get_models())

        return {
            'n': count,
            'ok': 1
        }


responder = Responce()



from mindsdb.api.mysql.mysql_proxy.controllers.session_controller import SessionController
from mindsdb.api.mysql.mysql_proxy.libs.constants.mysql import CHARSET_NUMBERS
from mindsdb.interfaces.model.model_interface import ModelInterface
from mindsdb.interfaces.database.integrations import IntegrationController
from mindsdb.interfaces.database.views import ViewController
from mindsdb.api.mysql.mysql_proxy.mysql_proxy import MysqlProxy


def empty_fn():
    pass


class Dummy:
    pass


class FakeMysqlProxy(MysqlProxy):
    def __init__(self, company_id, user_class):
        request = Dummy()
        client_address = ['', '']
        server = Dummy()
        server.connection_id = 0
        server.hook_before_handle = empty_fn
        server.original_model_interface = ModelInterface()
        server.original_integration_controller = IntegrationController()
        server.original_view_controller = ViewController()

        self.charset = 'utf8'
        self.charset_text_type = CHARSET_NUMBERS['utf8_general_ci']
        self.client_capabilities = None

        self.request = request
        self.client_address = client_address
        self.server = server

        self.session = SessionController(
            server=self.server,
            company_id=company_id,
            user_class=user_class
        )
        self.session.database = 'mindsdb'

    def is_cloud_connection(self):
        return {
            'is_cloud': False
        }



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.utilities import log

class Responce(Responder):
    when = {'saslStart': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        try:
            payload = query['payload'].decode()
            mechanism = query.get('mechanism', 'SCRAM-SHA-1')
            if mechanism == 'SCRAM-SHA-1':
                session.init_scram('sha1')
            elif mechanism == 'SCRAM-SHA-256':
                session.init_scram('sha256')
            responce = session.scram.process_client_first_message(payload)

            res = {
                'conversationId': 1,
                'done': False,
                'payload': responce.encode(),
                'ok': 1
            }
        except Exception as e:
            log.warning(str(e))
            res = {
                'errmsg': str(e),
                'ok': 0
            }
        return res


responder = Responce()



import json
from copy import deepcopy

import kafka

from mindsdb.interfaces.stream.base import StreamIntegration
import mindsdb.interfaces.storage.db as db
from mindsdb_streams import KafkaStream, StreamController, StreamLearningController


class KafkaConnectionChecker:
    def __init__(self, **params):
        self.connection_info = params['connection']

    def check_connection(self):
        try:
            client = kafka.KafkaClient(**self.connection_info)
        except Exception:
            return False
        else:
            client.close()
            return True


class Kafka(StreamIntegration, KafkaConnectionChecker):
    def __init__(self, config, name, db_info):
        self.connection_info = db_info['connection']

        # Back compatibility with initial API version
        self.control_stream = db_info.get('control_stream') or db_info.get('topic') or None
        if 'advanced' in db_info:
            self.connection_info['advanced'] = db_info['advanced']

        self.control_connection_info = deepcopy(self.connection_info)
        # don't need to read all records from the beginning of 'control stream'
        # since all active streams are saved in db. Use 'latest' auto_offset_reset for control stream
        if 'advanced' in self.control_connection_info:
            if 'consumer' in self.control_connection_info['advanced']:
                self.control_connection_info['advanced']['consumer']['auto_offset_reset'] = 'latest'

        StreamIntegration.__init__(
            self,
            config,
            name,
            control_stream=KafkaStream(self.control_stream, self.control_connection_info) if self.control_stream else None
        )

    def _make_stream(self, s: db.Stream):
        if s.learning_params and s.learning_threshold:
            learning_params = json.loads(s.learning_params) if isinstance(s.learning_params, str) else s.learning_params
            return StreamLearningController(
                s.name,
                s.predictor,
                learning_params,
                s.learning_threshold,
                stream_in=KafkaStream(s.stream_in, self.connection_info),
                stream_out=KafkaStream(s.stream_out, self.connection_info),
                in_thread=True
            )
        return StreamController(
            s.name,
            s.predictor,
            stream_in=KafkaStream(s.stream_in, self.connection_info),
            stream_out=KafkaStream(s.stream_out, self.connection_info),
            stream_anomaly=KafkaStream(s.anomaly_stream, self.connection_info) if s.anomaly_stream is not None else None,
            in_thread=True
        )



import base64

from mindsdb.api.mongo.classes.scram import Scram


class Session():
    def __init__(self, server_mindsdb_env):
        self.config = server_mindsdb_env['config']
        self.mindsdb_env = {'company_id': None}
        self.mindsdb_env.update(server_mindsdb_env)

    def init_scram(self, method):
        self.scram = Scram(method=method, get_salted_password=self.get_salted_password)

    def get_salted_password(self, username, method=None):
        real_user = self.config['api']['mongodb'].get('user', '')
        password = self.config['api']['mongodb'].get('password', '')
        if username != real_user:
            raise Exception(f'Wrong username {username}')

        salted_password = self.scram.salt_password(real_user, password)

        return base64.b64decode(self.scram.salt), salted_password



from mindsdb.api.mysql.mysql_proxy.mysql_proxy import MysqlProxy
from mindsdb.utilities.config import Config
from mindsdb.utilities.log import initialize_log


def start(verbose=False):
    config = Config()

    initialize_log(config, 'mysql', wrap_print=True)

    MysqlProxy.startProxy()



from pandas import DataFrame

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser.ast import Identifier

from mindsdb.integrations.libs.base_handler import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)


class ViewHandler(DatabaseHandler):
    """
    This handler handles views
    """

    name = 'views'

    def __init__(self, name=None, **kwargs):
        self.view_controller = kwargs['view_controller']
        super().__init__(name)

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the PostgreSQL database
        :return: success status and error message if error occurs
        """
        return StatusResponse(True)

    def native_query(self, query: ASTNode) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in PostgreSQL
        :return: returns the records from the current recordset
        """
        view_name = query.from_table.parts[-1]
        view_meta = self.view_controller.get(name=view_name)

        subquery_ast = parse_sql(view_meta['query'], dialect='mysql')
        if query.from_table.parts[-1] != view_name:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message=f"Query does not contain view name '{view_name}': {query}"
            )

        # set alias
        subquery_ast.alias = Identifier(view_name)
        query.from_table = subquery_ast

        return Response(
            RESPONSE_TYPE.QUERY,
            query=query
        )

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement with eliminated rows that dont satisfy the WHERE condition
        """
        return self.native_query(query)
        # query_str = self.renderer.get_string(query, with_failback=True)
        # return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        List all views
        """
        views = self.view_controller.get_all()
        result = []
        for view_name, view_meta in views.items():
            result.append({
                'table_name': view_meta['name'],
                'table_type': 'VIEW'
            })
        response = Response(
            RESPONSE_TYPE.TABLE,
            DataFrame(result)
        )
        return response

    def get_columns(self, table_name):
        # TODO
        query = f"""
            SELECT
                column_name as "Field",
                data_type as "Type"
            FROM
                information_schema.columns
            WHERE
                table_name = '{table_name}'
        """
        return self.native_query(query)



import pandas as pd
from sqlalchemy.types import (
    Integer, Float, Text
)
from mindsdb_sql.parser.ast import Insert, Identifier, CreateTable, TableColumn, DropTables

from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode

from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import TablesRow, TABLES_ROW_TYPE


class IntegrationDataNode(DataNode):
    type = 'integration'

    def __init__(self, integration_name, ds_type, integration_controller):
        self.integration_name = integration_name
        self.ds_type = ds_type
        self.integration_controller = integration_controller
        self.integration_handler = self.integration_controller.get_handler(self.integration_name)

    def get_type(self):
        return self.type

    def get_tables(self):
        response = self.integration_handler.get_tables()
        if response.type is RESPONSE_TYPE.TABLE:
            result_dict = response.data_frame.to_dict(orient='records')
            result = []
            for row in result_dict:
                result.append(TablesRow.from_dict(row))
            return result
        else:
            raise Exception(f"Can't get tables: {response.error_message}")

    def has_table(self, tableName):
        return True

    def get_table_columns(self, tableName):
        return []

    def create_table(self, table_name_parts, columns, data, is_replace=False, is_create=False):
        # is_create - create table
        # is_replace - drop table if exists
        # is_create==False and is_replace==False: just insert

        table_columns_meta = []
        table_columns = []
        for table in columns:
            for column in columns[table]:
                column_type = None
                for row in data:
                    column_value = row[table][column]
                    if isinstance(column_value, int):
                        column_type = Integer
                    elif isinstance(column_value, float):
                        column_type = Float
                    elif isinstance(column_value, str):
                        column_type = Text
                column_type = column_type or Text
                table_columns.append(
                    TableColumn(
                        name=column[-1],
                        type=column_type
                    )
                )
                table_columns_meta.append({
                    'table': table,
                    'name': column,
                    'type': column_type
                })

        if is_replace:
            # drop
            drop_ast = DropTables(
                tables=[Identifier(parts=table_name_parts)],
                if_exists=True
            )
            result = self.integration_handler.query(drop_ast)
            if result.type == RESPONSE_TYPE.ERROR:
                raise Exception(result.error_message)
            is_create = True

        if is_create:
            create_table_ast = CreateTable(
                name=Identifier(parts=table_name_parts),
                columns=table_columns,
                is_replace=True
            )

            result = self.integration_handler.query(create_table_ast)
            if result.type == RESPONSE_TYPE.ERROR:
                raise Exception(result.error_message)

        insert_columns = [Identifier(parts=[x['name'][-1]]) for x in table_columns_meta]
        formatted_data = []
        for row in data:
            new_row = []
            for column_meta in table_columns_meta:
                value = row[column_meta['table']][column_meta['name']]
                python_type = str
                if column_meta['type'] == Integer:
                    python_type = int
                elif column_meta['type'] == Float:
                    python_type = float

                try:
                    value = python_type(value) if value is not None else value
                except Exception:
                    pass
                new_row.append(value)
            formatted_data.append(new_row)

        insert_ast = Insert(
            table=Identifier(parts=table_name_parts),
            columns=insert_columns,
            values=formatted_data
        )

        result = self.integration_handler.query(insert_ast)
        if result.type == RESPONSE_TYPE.ERROR:
            raise Exception(result.error_message)


    def query(self, query):
        result = self.integration_handler.query(query)

        if result.type == RESPONSE_TYPE.ERROR:
            raise Exception(result.error_message)
        if result.type == RESPONSE_TYPE.QUERY:
            return result.query, None
        if result.type == RESPONSE_TYPE.OK:
            return

        df = result.data_frame
        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in df.dtypes.items()
        ]
        data = df.to_dict(orient='records')
        return data, columns_info



class DataNode:
    type = 'meta'

    def __init__(self):
        pass

    def get_type(self):
        return self.type

    def get_tables(self):
        pass

    def has_table(self, tableName):
        pass

    def get_table_columns(self, tableName):
        pass

    def query(self, table=None, columns=None, where=None, where_data=None, order_by=None, group_by=None, came_from=None):
        return []




from typing import List


class ANSWER_TYPE:
    __slots__ = ()
    TABLE = 'table'
    OK = 'ok'
    ERROR = 'error'


ANSWER_TYPE = ANSWER_TYPE()


class ExecuteAnswer:
    def __init__(self,
                 # TODO remove, is not used
                 answer_type: ANSWER_TYPE,
                 columns=None,
                 data: List[List] = None,
                 status: int = None,
                 state_track: List[List] = None,
                 error_code: int = None,
                 error_message: str = None,
                 ):
        self.columns = columns
        self.data = data
        self.status = status
        self.state_track = state_track
        self.error_code = error_code
        self.error_message = error_message



import inspect


class WithKWArgsWrapper(object):
    """ Change default values of arguments in all methods in given instance.

    Attributes:
        original_instance: instance of wrapped class
        **kwargs: arguments key/values
    """

    def __init__(self, original_instance: object, **kwargs):
        self.default_kwargs = kwargs
        self.original_instance = original_instance
        self.wrapped_methods = {}
        for method_name, method in inspect.getmembers(original_instance, inspect.ismethod):
            method_meta = inspect.getfullargspec(method)
            for arg_name in kwargs:
                wrapped_method_meta = {
                    'args': [],
                    'index': []
                }
                if arg_name in method_meta.args:
                    wrapped_method_meta['args'].append(arg_name)
                    wrapped_method_meta['index'].append(method_meta.args.index(arg_name) - 1)
                elif method_meta.varkw == 'kwargs':
                    wrapped_method_meta['args'].append(arg_name)
                    wrapped_method_meta['index'].append(None)
                if len(wrapped_method_meta['args']) > 0:
                    self.wrapped_methods[method_name] = wrapped_method_meta

    def __getattr__(self, method_name: str):
        def wrapper(*args, **kwargs):
            method = getattr(self.original_instance, method_name)
            if method_name in self.wrapped_methods:
                wrapped_args_names = self.wrapped_methods[method_name]['args']
                wrapped_args_indexes = self.wrapped_methods[method_name]['index']
                for i, arg_name in enumerate(wrapped_args_names):
                    if wrapped_args_indexes[i] is not None and wrapped_args_indexes[i] < len(args):
                        continue
                    if arg_name not in kwargs:
                        kwargs[arg_name] = self.default_kwargs[arg_name]
            return method(*args, **kwargs)
        return wrapper

    @staticmethod
    def _test():
        class T:
            def one(self, *args, **kwargs):
                print(f"test one: {kwargs.get('test')}")
                return kwargs.get('test')

            def two(self, test=1):
                print(f'test two: {test}')
                return test

            def three(self, x, test=1):
                print(f'test three: {test}')
                return test

            def four(self, x, test=1, y='y'):
                print(f'test four: {test}')
                return test

        t = WithKWArgsWrapper(T(), test='x')

        assert t.one() == 'x'
        assert t.one(test=0) == 0
        assert t.one(test=None) is None

        assert t.two(2) == 2
        assert t.two() == 'x'
        assert t.two(test=0) == 0

        assert t.three(3) == 'x'
        assert t.three(3, 0) == 0
        assert t.three(3, test=0) == 0

        assert t.four(4) == 'x'
        assert t.four(4, 0, 4) == 0
        assert t.four(4, y=4, test=0) == 0
        assert t.four(4, test=0, y=4) == 0
        assert t.four(4, y=4) == 'x'


if __name__ == "__main__":
    WithKWArgsWrapper._test()



"""integration-args

Revision ID: 999bceb904df
Revises: d74c189b87e6
Create Date: 2022-07-08 10:58:19.822618

"""
import json

from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import text

import mindsdb.interfaces.storage.db


# revision identifiers, used by Alembic.
revision = '999bceb904df'
down_revision = 'd74c189b87e6'
branch_labels = None
depends_on = None


def upgrade():
    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)

    with op.batch_alter_table('integration', schema=None) as batch_op:
        batch_op.add_column(sa.Column('engine', sa.String()))

    integrations = conn.execute('''
        select id, name, data from integration
    ''').fetchall()

    for row in integrations:
        try:
            data = json.loads(row['data'])
        except Exception:
            data = {}

        if 'test' in data:
            del data['test']
        if 'publish' in data:
            del data['publish']
        if 'enabled' in data:
            del data['enabled']
        if 'database_name' in data:
            if row['name'] is None:
                row['name'] = data['database_name']
            del data['database_name']
        integration_type = data.get('type')
        if integration_type is None:
            if row['name'] == 'files':
                integration_type = 'files'
            if row['name'] == 'views':
                integration_type = 'views'
        if 'type' in data:
            del data['type']

        conn.execute(
            text("""
                update integration
                set engine = :integration_type,
                    data = :integration_data
                where id = :integration_id
            """), {
                'integration_type': integration_type,
                'integration_data': json.dumps(data),
                'integration_id': row['id']
            }
        )

    session.commit()


def downgrade():
    conn = op.get_bind()
    session = sa.orm.Session(bind=conn)

    integrations = conn.execute('''
        select id, name, type, data from integration
    ''').fetchall()

    for row in integrations:
        try:
            data = json.loads(row['data'])
        except Exception:
            data = {}
        if row['engine'] is not None:
            data['type'] = row['engine']

        conn.execute(
            text("""
                update integration
                set data = :integration_data
                where id = :integration_id
            """), {
                'integration_data': json.dumps(data),
                'integration_id': row['id']
            }
        )

    with op.batch_alter_table('integration', schema=None) as batch_op:
        batch_op.drop_column('engine')

    session.commit()
